
/*******************************************************************************

    demo1.c
    18 Apr 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

This is a minimalist demonstration of how to use the igig wrapper to (try to)
find a fixed point.  The function is 4D, but is effectively four independent 1D 
functions, which we solve simultaneously.

*******************************************************************************/

#include "igig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/*******************************************************************************
This 4D function is drawn from the example non-trivial involutions on the unit
interval given in "Fixed Points" (Yu A Shashkin, AMS, 1991) as Problem 4 of 
Chapter 1 (pp 3-4).  These all have unique fixed-points (in the interval), and 
we've extended the domain to R while preserving continuity and the range (we can
do this without problems since all the functions have f(0)=1 & f(1)=0).  The 
functions and their fixed-points are:
  1-x                  1/2            ~=  0.5
  (1-x)/(1+x)          sqrt(2) - 1    ~=  0.414213562
  sqrt(1-x^2)          1/sqrt(2)      ~=  0.707106781
  1-sqrt(1-(x-1)^2)    1 - 1/sqrt(2)  ~=  0.292893219
*******************************************************************************/

igBOOL f1Apply(double *x, double *y)
  {
  if      (x[0] <= 0.0) { y[0] = 1.0; }
  else if (x[0] >= 1.0) { y[0] = 0.0; }
  else                  { y[0] = 1 - x[0]; }

  if      (x[1] <= 0.0) { y[1] = 1.0; }
  else if (x[1] >= 1.0) { y[1] = 0.0; }
  else                  { y[1] = (1 - x[1])/(1 + x[1]); }

  if      (x[2] <= 0.0) { y[2] = 1.0; }
  else if (x[2] >= 1.0) { y[2] = 0.0; }
  else                  { y[2] = sqrt(1 - x[2]*x[2]); }

  if      (x[3] <= 0.0) { y[3] = 1.0; }
  else if (x[3] >= 1.0) { y[3] = 0.0; }
  else                  { y[3] = 1 - sqrt(1 - (x[3]-1)*(x[3]-1)); }

  return igTRUE;
  }

int main(int argc, char *argv[])
  {
  igList *igl;
  igFPSTAT igfps;
  int i;

  igigSetup();

  igDim = 4;
  igFunction = *f1Apply;
  igigDmin = 0.0;
  igigDmax = 1.0;

  igl = igNewList(igDim);

  igfps = igigFindFP(igl, 0U);

  if (igfps == igFail)
    { printf("Unable to find any (approx) fixed point\n"); }
  else
    {
    if (igfps == igFP)
      { printf("Found a fixed point, norm = %0.15g\n", igl->dn[igl->size-1]); }
    else
      { printf("Best approx'n found, norm = %0.15g\n", igl->dn[igl->size-1]); }

    for (i = 0; i < igDim; i++)
      { printf("  X[%d] = %0.15g  Y[%d] = %0.15g\n",
                      i, igl->dx[igl->size-1][i], i, igl->dy[igl->size-1][i]); }
    }

  free(igl);
  exit(EXIT_SUCCESS);
  }


