
/*******************************************************************************

    demo2.c
    4 May 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

As demo1.c, but somewhat less minimalist.  We simultaneously solve three 1D
problems via a 3D problem, but here the 1D functions are 'mixed-up' in the 3D
function.

*******************************************************************************/

#include "igig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/*******************************************************************************
Given a $mu > 1$, let $\lambda = mu/e^{mu}$.  Then $f(x) = \lambda e^x has two
fixed points, one at $mu$ and one in $(0,1)$.  We build a (randomish) 3D 
function from 3 of these, with a little bit of co-mingling (via the 3x3 matrix 
$M$).  The (vector) function is Y = L.e^X + M.(X - L.e^X).  This has (at least)
2^3 = 8 FPs, but the ones at the $mu$ values are hard to find, since they're 
repulsive and the exponential function zooms off to infinity.
*******************************************************************************/

static double mumax = 4.0;
static double mu[3];
static double lamb[3];

static double Mscale = 1.0/8.0;
static double M[3][3];

static int f2call;

void f2Init(void)
  {
  int i,j;

  /* NB: 0 < (rand() + 0.5)/(RAND_MAX + 1.0) < 1, so 1 < mu[i] < mumax.  */

  mu[0] = 1.0 + (mumax - 1.0)*((double)rand() + 0.5)/((double)RAND_MAX + 1.0);
  mu[1] = 1.0 + (mumax - 1.0)*((double)rand() + 0.5)/((double)RAND_MAX + 1.0);
  mu[2] = 1.0 + (mumax - 1.0)*((double)rand() + 0.5)/((double)RAND_MAX + 1.0);

  lamb[0] = mu[0] / exp(mu[0]);
  lamb[1] = mu[1] / exp(mu[1]);
  lamb[2] = mu[2] / exp(mu[2]);

  for (i = 0; i < 3; i++)
    {
    for (j = 0; j < 3; j++)
      { M[i][j] = igNormal(0.0, Mscale); }
    }

  f2call = 0;
  }

/*******************************************************************************
To prevent numerical problems we 'clip' the output values if the input values
are outside the possible domain of the FPs of (0,mumax) + a bit of slack.  We 
make this interval wider than (igigDmin,igigDmax) to keep things 'sensible' (?),
and do the clipping in such a way as to preserve continuity without introducing
'false' FPs.  In fact, we only need to deal with the >mumax x values; at mumax+3
f(x) > x, and we 'clip' f(x) to run parallel to the y=x line.
*******************************************************************************/

igBOOL f2Apply(double *x, double *y)
  {
  int i;
  double fx[3], diff[3];

  for (i = 0; i < 3; i++)
    {
    if (x[i] >= mumax + 3.0) 
      { fx[i] = lamb[i] * exp(mumax + 3.0) + (x[i] - (mumax + 3.0)); }
    else
      { fx[i] = lamb[i] * exp(x[i]); }

    diff[i] = x[i] - fx[i];
    }

  for (i = 0; i < 3; i++)
    { y[i] = fx[i] + M[i][0]*diff[0] + M[i][1]*diff[1] + M[i][2]*diff[2]; }

  f2call++;
  return igTRUE;
  }

int main(int argc, char *argv[])
  {
  igList *igl;
  igFPSTAT igfps;
  int i;

  /* Initialisation & setup ... */

  igigSetup();

  f2Init();
  igDim = 3;
  igFunction = *f2Apply;

  /* The FPs are all in (0,mumax), so we set our domain to this + a bit. */

  igigDmin = -2.0;
  igigDmax = mumax + 2.0;

  /* Any non-default set-up ... */

  igTol = 1.0E-12;

  /* Dump the function's details ... */

  printf("\n");
  printf("  mu[] = %0.15g  %0.15g  %0.15g\n", mu[0], mu[1], mu[2]);
  printf("lamb[] = %0.15g  %0.15g  %0.15g\n", lamb[0], lamb[1], lamb[2]);
  printf(" M[][] = %0.15g  %0.15g  %0.15g\n", M[0][0], M[0][1], M[0][2]);
  printf("       = %0.15g  %0.15g  %0.15g\n", M[1][0], M[1][1], M[1][2]);
  printf("       = %0.15g  %0.15g  %0.15g\n", M[2][0], M[2][1], M[2][2]);

  /* Create results list, try to find FP & process result ... */

  igl = igNewList(igDim);
  igfps = igigFindFP(igl, 0U);

  printf("\n");
  if (igfps == igFail)
    { printf("Unable to find any (approx) fixed point\n"); }
  else
    {
    if (igfps == igFP)
      { printf("Found a fixed point, norm = %0.15g\n", igl->dn[igl->size-1]); }
    else
      { printf("Best approx'n found, norm = %0.15g\n", igl->dn[igl->size-1]); }

    for (i = 0; i < igDim; i++)
      { printf("  X[%d] = %0.15g  Y[%d] = %0.15g\n",
                      i, igl->dx[igl->size-1][i], i, igl->dy[igl->size-1][i]); }
    }

  printf("\n");
  printf("Number of function calls = %d\n", f2call);
  printf("\n");

  free(igl);

  exit(EXIT_SUCCESS);
  }


