
/*******************************************************************************

    example1.c
    2 Jun 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

This example demonstrates how fixed points can be used to find roots of 
polynomials.  If the polynomial is f(x) = a_n x^n + ..., then the fixed-point
problem is g(x) = x + f(x); so g(x) = x iff f(x) = 0.  Note that the norm 
tolerance used to test for a fixed-point is applied to g(x), so we have
||x - g(x)|| = ||x - (x + f(x)|| = ||f(x)||.

Note that the polynomials are unbounded as |x| increases.  In many cases the
isfinite() calls catch NAN/INF values and cause the igImitate() et al routines 
to return igFail well before they reach their iteration limits.  If not, they 
run all the way, and 'waste' time.  Is this OK, or should we implement some form
'clipping'?  Formally, the IG algorithm only 'works' for bounded & continuous 
functions.

*******************************************************************************/

#include "igfp.h"      /* includes ig.h & iglist.h */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*******************************************************************************
All of the parameters for this example programme are accessible anywhere within 
this file, and they are passed as parameters to the IG library as required.  Any
features of the library not used, or not settable by the user, as set to default
values (either explicitly or implicitly).
*******************************************************************************/

unsigned int seed;
int deg;
int dim;
int fmode;
double rmin, rmax;
int meth;
int xmode;
int att;
int norm;
double eps;
double zero;
int lim;
int piv;
int lhpivot;

/*******************************************************************************
This is the stuff we need to create & use our example function ...
*******************************************************************************/

double *coeff = NULL;      /* the coefficient vector */

void f1Init(void)
  {
  int i;
  coeff = igGetVect(deg+1);
  for (i = 0; i <= deg; i++)
    { coeff[i] = (double)(rand()%(2*fmode + 1) - fmode); }
  }

/* Returns true if coeff[i] is 0 for all the non-constant (ie, ex x^0) terms.
Although the coeff[i] are doubles, the values are 'integral'. */

igBOOL f1Triv(void)
  {
  int i;
  for (i = 1; i <= deg; i++)
    { if (coeff[i] != 0.0) { return igFALSE; } }
  return igTRUE;
  }

void f1Free(void)
  { if (coeff != NULL) { free(coeff); } }

/* NB: Both X & Y are 1-dimensional, and f(X) = X + polynomial. */

igBOOL f1Apply(double *x, double *y)
  {
  int i;
  double tmp;

  y[0] = x[0];

  tmp = coeff[deg] * x[0];
  for (i = deg-1; i > 0; i--)
    { tmp = (tmp + coeff[i]) * x[0]; }
  tmp += coeff[0];

  y[0] += tmp;

  return igTRUE;
  }

void f1Print()
  {
  int i;

  printf("f(x)  = ");
  if (coeff == NULL)
    { printf(" NULL\n"); }
  else
    {
    for (i = deg; i >= 0; i--)
      {
      if (coeff[i] < 0.0)
        { 
        printf(" - %0.15g", -coeff[i]);
        if      (i == 0) { ; }
        else if (i == 1) { printf(" x"); }
        else             { printf(" x^%d", i); }
        }
      else if (coeff[i] > 0.0)
        { 
        printf(" + %0.15g", coeff[i]);
        if      (i == 0) { ; }
        else if (i == 1) { printf(" x"); }
        else             { printf(" x^%d", i); }
        }
      }
    printf("\n");
    }
  }

/*******************************************************************************
A quick kludge to dump an igList, in the 1-dimensional case.
*******************************************************************************/

void dumpIGL(igList *igl)
  {
  int i;
  printf("  dim = %d  alloc = %d  size = %d\n", 
                                               igl->dim, igl->alloc, igl->size);
  for (i = 0; i < igl->size; i++)
    { printf("  x = %0.15g  y = %0.15g  n = %0.15g\n", 
                                    igl->dx[i][0], igl->dy[i][0], igl->dn[i]); }
  }

/*******************************************************************************
These are the utility programmes we need ...
*******************************************************************************/

/* This is for user (ie, parameter) errors, not IG library errors. */

void errorExit(char *s)
  {
  fprintf(stderr, "ERROR: %s\n", s);
  exit(EXIT_FAILURE);
  }

/* Print the opening banner message. */

void printBanner(void)
  {
  printf("Find polynomial roots, using fixed points & imitation games\n");
  printf("%s is Copyright (C) 2011, The Gambit Project\n", IG);
  printf("This is free software, distributed under the GNU GPL\n");
  }

/* Print the help message. */

void printHelp(char *s)
  {
  printf("\n");
  printf("Usage: %s [OPTIONS]\n", s);
  printf("Options:\n");
  printf("  --attempt A     Make A (> 0) attempts to find a root\n");
  printf("  --degree A      Set the degree to A (1 to 8)\n");
  printf("  --eps A         Set the norm tolerance to A (> 0.0)\n");
  printf("  --fmode A       Set the function mode to A (1 to 8)\n");
  printf("  -(-h)elp        Print this help screen and then terminate\n");
  printf("  --lh A          Set Lemke-Howson pivot mode to A (0, 1 or 2)\n");
  printf("  --limit A       Set iteration/game-size limit to A (> 1)\n");
  printf("  --method A      Set the solution method to A (1 or 2)\n");
  printf("  --norm A        Set the norm to use to A (-1, 1 or 2)\n");
  printf("  --pivot A       Set maximum pivots per Lemke-Howson to A (> 1)\n");
  printf("  --region A B    Set region of interest to the interval [A,B]\n");
  printf("  --seed A        Seed the random number generator with A (> 0)\n");
  printf("  --xmode A       Set the initial X mode to A (> 0)\n");
  printf("  --zero A        Set the floating-point tolerance to A (> 0.0)\n");
  printf("\n");
  }

/* Print the parameters. */

void printParam(void)
  {
  /* IGFP stuff ... */

  printf("dim = %d  tol = %0.15g  zero = %0.15g  pivot = %d", 
                                                igDim, igTol, igZero, igPLimit);
  if (lhpivot == 2) 
    { printf("  lh = best\n"); }
  else if (lhpivot == 1) 
    { printf("  lh = norm\n"); }
  else  
    { printf("  lh = last\n"); }

  /* IG & function stuff ... */

  printf("seed = %u  deg = %d  fmode = %d  region = [%0.15g,%0.15g]\n", 
                                          igRandomSeed, deg, fmode, rmin, rmax);

  /* Programme stuff ... */

  printf("method = %d  xmode = %d  attempt = %d", meth, xmode, att);
  if (norm == -1)
    { printf("  norm = Linf"); }
  else if (norm == 1)
    { printf("  norm = L1"); }
  else
    { printf("  norm = L2"); }
  printf("  limit = %d\n", lim);
  }

/* Set all the parameters to their initial default values. */

void initParam(void)
  {
  seed = 0;            /* generate a (random) seed internally */
  deg = 3;             /* degree of the polynomial */
  dim = 1;             /* dimension of the problem space (fixed) */
  fmode = 4;           /* function mode, aka polynomial coefficients */
  rmin = 1.0;          /* region min & max are to-be-determined, */
  rmax = 0.0;          /*   so we set min > max */
  meth = 1;            /* use the imitation game method */
  xmode = 1000;        /* initial X mode is best of 1000 */
  att = 4;             /* make 4 attempts */
  norm = -1;           /* use Linfinity norm */
  eps = 1.0E-5;        /* epsilon for testing norm */
  zero = 1.0E-14;      /* floating-pt zero tolerance */
  lim = 150;           /* max imitation game size / no. of iterations */
  piv = 50;            /* max pivots per Lemke-Howson call */
  lhpivot = 0;         /* use last X as initial pivot */
  }

/* Process all the arguments and set the parameters as directed. */

void procArg(int argc, char *argv[])
  {
  int i = 1;

  while (i < argc)
    {
    /* These options will terminate the programme */

    if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0) 
      { printHelp(argv[0]);  exit(EXIT_SUCCESS); }

    /* Options with one parameter */

    if (strcmp(argv[i], "--attempt") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--attempt\""); }
      att = atoi(argv[i]);
      if (att < 1)
        { errorExit("number of attempts must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--degree") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--degree\""); }
      deg = atoi(argv[i]);
      if (deg < 1 || deg > 8)
        { errorExit("degree must be in the range [1,8]"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--eps") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--eps\""); }
      eps = atof(argv[i]);
      if (eps <= 0.0)
        { errorExit("epsilon value must be strictly positive"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--fmode") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--fmode\""); }
      fmode = atoi(argv[i]);
      if (fmode < 1 || fmode > 8)
        { errorExit("function mode must be in range [1,8]"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--lh") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--lh\""); }
      lhpivot = atoi(argv[i]);
      if (lhpivot < 0 || lhpivot > 2)
        { errorExit("LH pivoting choice method must be 0, 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--limit") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--limit\""); }
      lim = atoi(argv[i]);
      if (lim < 2)
        { errorExit("game-size/iteration limit must be at least 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--method") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--method\""); }
      meth = atoi(argv[i]);
      if (meth < 1 || meth > 2)
        { errorExit("solution method must be 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--norm") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--norm\""); }
      norm = atoi(argv[i]);
      if (norm != -1 && norm != 1 && norm != 2)
        { errorExit("norm must be -1 (inf), 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--pivot") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--pivot\""); }
      piv = atoi(argv[i]);
      if (piv < 2)
        { errorExit("pivot limit must be at least 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--seed") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--seed\""); }
      seed = (unsigned int)strtoul(argv[i], NULL, 10);
      if (seed < 1)
        { errorExit("seed must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--xmode") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--xmode\""); }
      xmode = atoi(argv[i]);
      if (xmode < 1)
        { errorExit("initial X mode must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--zero") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--zero\""); }
      zero = atof(argv[i]);
      if (zero <= 0.0)
        { errorExit("zero value must be strictly positive"); }
      i++;
      continue;
      }

    /* Options with two parameters */

    if (strcmp(argv[i], "--region") == 0)
      {
      i += 2;
      if (!(i < argc))
        { errorExit("missing arguments for \"--region\""); }
      rmin = atof(argv[i-1]);
      rmax = atof(argv[i]);
      if (!(isfinite(rmin) && isfinite(rmax) && rmin < rmax))
        { errorExit("region must be a non-empty bounded interval"); }
      i++;
      continue;
      }

    /* If we fall through to here, the argument was unrecognised and it's an
    error.  At some point in the future, unrecognised arguments might be, eg,
    treated as filenames (if they don't start with "-"), and processed here. */

    fprintf(stderr, "ERROR: unrecognised argument, \"%s\"\n", argv[i]);
    exit(EXIT_FAILURE);
    }
  }

/* Check the parameters & do some setup/initialisation of IG library. */

void chkSetup(void)
  {
  /* A simple bound, based on the Rouche theorem, on the absolute value of the
  real roots is given by R = 1 + max{|a_0|, ..., |a_n-1|}/|a_n|.  In our case 
  the a_i are 'integers' in the range [-fmode,+fmode], so the largest possible
  value of R = 1 + fmode (plus a little bit extra). */

  if (rmin > rmax)       /* if values not explicitly set */
    {
    rmax = (double)(2+fmode);
    rmin = -rmax;
    }

  /* Seed the random number generator and set all the igfp stuff, except for the
  function. */

  igSeedRand(seed);

  igDim = dim;
  igTol = eps;
  igZero = zero;
  igPLimit = piv;

  if      (lhpivot == 2) { igLHPivot = igLHbest; }
  else if (lhpivot == 1) { igLHPivot = igLHnorm; }
  else                   { igLHPivot = igLHlast; }

  if      (norm == -1) { igNorm = *igLinfNorm; }
  else if (norm == 1)  { igNorm = *igL1Norm; }
  else                 { igNorm = *igL2Norm; }
  }

/*******************************************************************************
The entry point and top-level wrapper for the whole shebang ...
*******************************************************************************/

int main(int argc, char *argv[])
  {
  int a, i, j;
  double n;
  igList *igl = NULL;
  igFPSTAT igfps;

  initParam();
  procArg(argc, argv);
  chkSetup();

  printf("\n");
  printBanner();
  printf("\n");
  printParam();

  /* Setup the function to use. */

  f1Init();
  while (f1Triv()) { f1Free();  f1Init(); }
  printf("\n");
  f1Print();
  igFunction = *f1Apply;

  /* Each attempt uses a new igList, but the same function. */

  for (a = 1; a <= att; a++)
    {
    printf("\n");
    printf("Attempt #%d ...\n", a);

    if (igl != NULL) { igFreeList(igl); }
    igl = igNewList(igDim);

    /* Generate a starting X/Y/norm triple. */

    igfps = igGuess(igl, xmode, rmin, rmax);
    if (igfps == igFP)
      {
      printf("  ROOT, guessed, |list| = %d\n", igl->size);
      printf("    X = %0.15g\n", igl->dx[igl->size-1][0]);
      printf("    Y = %0.15g\n", 
                             igl->dy[igl->size-1][0] - igl->dx[igl->size-1][0]);
      continue;
      }
    else if (igfps == igOK)
      { ; }
    else
      {
      printf("  guessing phase failed, |list| = %d\n", igl->size);
      continue;
      }

    /* Run the imitate or iterate FP finder. */

    if (meth == 1)     /* use imitation games */
      {
      /* At this point the list contains precisely 1 point, so the first IG game
      will have size 1x1.  The lim parameter is at least 2 and, to ensure that 
      the largest IG has size at most lim, we pass lim as the "cnt" argument to
      igImitate(). */

      igfps = igImitate(igl, lim, 0U);
      if (igfps == igFP)
        {
        printf("  ROOT, via imitation, |list| = %d\n", igl->size);
        printf("    X = %0.15g\n", igl->dx[igl->size-1][0]);
        printf("    Y = %0.15g\n", 
                             igl->dy[igl->size-1][0] - igl->dx[igl->size-1][0]);
        continue;
        }
      else if (igfps == igOK)
        { ; }
      else
        {
        printf("  imitation phase failed, |list| = %d\n", igl->size);
        continue;
        }
      }
    else               /* use iteration */
      {
      igfps = igIterate(igl, lim);
      if (igfps == igFP)
        {
        printf("  ROOT, via iteration, |list| = %d\n", igl->size);
        printf("    X = %0.15g\n", igl->dx[igl->size-1][0]);
        printf("    Y = %0.15g\n", 
                             igl->dy[igl->size-1][0] - igl->dx[igl->size-1][0]);
        continue;
        }
      else if (igfps == igOK)
        { ; }
      else
        {
        printf("  iteration phase failed, |list| = %d\n", igl->size);
        continue;
        }
      }

    /* If we fall through to here there haven't been any problems but we haven't
    found a within-tolerance FP.  We have however appended one or more X/Y/norm
    tuples to the list, so we dump the best approx'n found. */

/*
    dumpIGL(igl);
*/

    n = igl->dn[0];
    j = 0;
    for (i = 1; i < igl->size; i++)
      { if (igl->dn[i] < n) { n = igl->dn[i];  j = i; } }

    printf("  best approximate root, |list| = %d\n", igl->size);
    printf("    X[%d] = %0.15g\n", j, igl->dx[j][0]);
    printf("    Y[%d] = %0.15g\n", j, igl->dy[j][0] - igl->dx[j][0]);
    }

  f1Free();
  if (igl != NULL) { igFreeList(igl); }

  printf("\n");
  exit(EXIT_SUCCESS);
  }


