
/*******************************************************************************

    example2a.c
    2 Jun 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

This is the same as example2, but has been extended to use the GSL software to 
polish the best approximate eigenvector, in a similar fashion to example1a.c.

NB: All of ig{Guess|Iterate|Imitate|Polish}() have the property that, if they 
update the igList and return igFP, then the function's lambda is correctly set 
for the last entry in the updated igList.  This is also true for a return of 
igOK for igPolish(), but is *not* true for the others.  Note that igImitate()
can add multiple values to the igList, but the others add at most one value.

*******************************************************************************/

#include "iggsl.h"     /* includes igfp.h (and so ig.h & iglist.h) */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsl/gsl_version.h>

/*******************************************************************************
All of the parameters for this example programme are accessible anywhere within 
this file, and they are passed as parameters to the IG library as required.  Any
features of the library not used, or not settable by the user, as set to default
values (either explicitly or implicitly).
*******************************************************************************/

unsigned int seed;
int dim;
int fmode;
int meth;
int att;
int norm;
double eps;
double zero;
int lim;
int piv;
int lhpivot;
int gsliter;

/*******************************************************************************
This is the stuff we need to create & use our example function ...
*******************************************************************************/

double **A;          /* the array */
double lambda;       /* the latest value of |Ax| (ie, last apply() call) */
int callCnt;         /* number of function calls (since last reset) */

/* The dimension is at least 2, and the array entries are drawn uniformly at 
random from the integers (as doubles) in [0,fmode]. */

void f1Init(void)
  {
  int i, j;

  if ((A = (double**)malloc(dim * sizeof(double*))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in f1Init()"); }
  for (i = 0; i < dim; i++)
    {
    A[i] = igGetVect(dim);
    for (j = 0; j < dim; j++)
      { A[i][j] = (double)(rand()%(fmode+1)); }
    }

  lambda = -1.0;     /* the "not calculated" sentinal */
  }

void f1Free(void)
  { 
  int i;
  if (A != NULL) 
    { for (i = 0; i < dim; i++) { free(A[i]); } }
  free(A);
  }

/* All the A[i][j] & x[i] are (or should be) non-negative, so the y[i] & lambda 
will be too.  The A & x entries are also limited in size, so there's no danger 
of an overflow.  However, we need to check that the |Ax| value (ie, lambda) is 
not too small before we divide by it.  If it is, we return FALSE.  Note that 
|Ax| is only an eigenvalue when (a) it's 0 and x[] is non-zero, or (b) x[] is 
stochastic and we're at a fixed point. */

igBOOL f1Apply(double *x, double *y)
  {
  int i, j;
  double c;

  callCnt++;
  lambda = 0.0;
  
  for (i = 0; i < dim; i++)
    {
    c = 0.0;
    for (j = 0; j < dim; j++) { c += A[i][j] * x[j]; }
    y[i] = c;
    lambda += c;
    }
 
  if (lambda < zero) { return igFALSE; }

  for (i = 0; i < dim; i++) { y[i] /= lambda; }

  return igTRUE;
  }

/* Pretty-print the matrix A, for the user.  Intended for small matrices with
integer entries. */

void f1Print()
  {
  int i, j;

  printf("A =");
  if (A == NULL)
    { printf(" NULL\n"); }
  else
    {
    for (i = 0; i < dim; i++)
      {
      if (i > 0) { printf("   "); }
      for (j = 0; j < dim; j++) { printf(" %0.15g", A[i][j]); }
      printf("\n");
      }
    }
  }

/*******************************************************************************
These are the utility programmes we need ...
*******************************************************************************/

/* This is for user (ie, parameter) errors, not IG library errors. */

void errorExit(char *s)
  {
  fprintf(stderr, "ERROR: %s\n", s);
  exit(EXIT_FAILURE);
  }

/* Print the opening banner message. */

void printBanner(void)
  {
  printf("Find eigenvectors & eigenvalues,"
                                     " using fixed points & imitation games\n");
  printf("%s is Copyright (C) 2011, The Gambit Project\n", IG);
  printf("Incorporating GSL version %s\n", GSL_VERSION);
  printf("This is free software, distributed under the GNU GPL\n");
  }

/* Print the help message. */

void printHelp(char *s)
  {
  printf("\n");
  printf("Usage: %s [OPTIONS]\n", s);
  printf("Options:\n");
  printf("  --attempt A     Make A (> 0) attempts to find a root\n");
  printf("  --dim A         Set the dimension to A (2 to 8)\n");
  printf("  --eps A         Set the norm tolerance to A (> 0.0)\n");
  printf("  --fmode A       Set the function mode to A (1 to 25)\n");
  printf("  -(-h)elp        Print this help screen and then terminate\n");
  printf("  --iter A        Set the GSL iteration limit to A (> 0)\n");
  printf("  --lh A          Set Lemke-Howson pivot mode to A (0, 1 or 2)\n");
  printf("  --limit A       Set iteration/game-size limit to A (> 1)\n");
  printf("  --method A      Set the solution method to A (1 or 2)\n");
  printf("  --norm A        Set the norm to use to A (-1, 1 or 2)\n");
  printf("  --pivot A       Set maximum pivots per Lemke-Howson to A (> 1)\n");
  printf("  --seed A        Seed the random number generator with A (> 0)\n");
  printf("  --zero A        Set the floating-point tolerance to A (> 0.0)\n");
  printf("\n");
  }

/* Print the parameters. */

void printParam(void)
  {
  /* IGFP stuff ... */

  printf("dim = %d  tol = %0.15g  zero = %0.15g  pivot = %d", 
                                                igDim, igTol, igZero, igPLimit);
  if (lhpivot == 2) 
    { printf("  lh = best\n"); }
  else if (lhpivot == 1) 
    { printf("  lh = norm\n"); }
  else  
    { printf("  lh = last\n"); }

  /* IG & function stuff & GSL stuff ... */

  printf("seed = %u  fmode = %d  iter = %d\n", igRandomSeed, fmode, gsliter);

  /* Programme stuff ... */

  if (meth == 1)
    { printf("method = imitate"); }
  else
    { printf("method = iterate"); }
  printf("  attempt = %d", att);
  if (norm == -1)
    { printf("  norm = Linf"); }
  else if (norm == 1)
    { printf("  norm = L1"); }
  else
    { printf("  norm = L2"); }
  printf("  limit = %d\n", lim);
  }

/* Set all the parameters to their initial default values. */

void initParam(void)
  {
  seed = 0;            /* generate a (random) seed internally */
  dim = 3;             /* dimension of the problem space */
  fmode = 2;           /* function mode, aka max array entry */
  meth = 1;            /* use the imitation game method */
  att = 3;             /* make three attempts */
  norm = -1;           /* use Linfinity norm */
  eps = 1.0E-9;        /* epsilon for testing norm */
  zero = 4.0E-15;      /* floating-pt zero tolerance */
  lim = 100;           /* max imitation game size / no. of iterations */
  piv = 50;            /* max pivots per Lemke-Howson call */
  lhpivot = 0;         /* use last X as initial pivot */
  gsliter = 100;       /* limit on GSL polishing steps */
  }

/* Process all the arguments and set the parameters as directed. */

void procArg(int argc, char *argv[])
  {
  int i = 1;

  while (i < argc)
    {
    /* These options will terminate the programme */

    if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0) 
      { printHelp(argv[0]);  exit(EXIT_SUCCESS); }

    /* Options with one parameter */

    if (strcmp(argv[i], "--attempt") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--attempt\""); }
      att = atoi(argv[i]);
      if (att < 1)
        { errorExit("number of attempts must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--dim") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--dim\""); }
      dim = atoi(argv[i]);
      if (dim < 2 || dim > 8)
        { errorExit("dimension must be in the range [2,8]"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--eps") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--eps\""); }
      eps = atof(argv[i]);
      if (eps <= 0.0)
        { errorExit("epsilon value must be strictly positive"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--fmode") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--fmode\""); }
      fmode = atoi(argv[i]);
      if (fmode < 1 || fmode > 25)
        { errorExit("function mode must be in range [1,25]"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--iter") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--iter\""); }
      gsliter = atoi(argv[i]);
      if (gsliter < 1)
        { errorExit("GSL iteration limit must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--lh") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--lh\""); }
      lhpivot = atoi(argv[i]);
      if (lhpivot < 0 || lhpivot > 2)
        { errorExit("LH pivoting choice method must be 0, 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--limit") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--limit\""); }
      lim = atoi(argv[i]);
      if (lim < 2)
        { errorExit("game-size/iteration limit must be at least 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--method") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--method\""); }
      meth = atoi(argv[i]);
      if (meth < 1 || meth > 2)
        { errorExit("solution method must be 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--norm") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--norm\""); }
      norm = atoi(argv[i]);
      if (norm != -1 && norm != 1 && norm != 2)
        { errorExit("norm must be -1 (inf), 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--pivot") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--pivot\""); }
      piv = atoi(argv[i]);
      if (piv < 2)
        { errorExit("pivot limit must be at least 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--seed") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--seed\""); }
      seed = (unsigned int)strtoul(argv[i], NULL, 10);
      if (seed < 1)
        { errorExit("seed must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--zero") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--zero\""); }
      zero = atof(argv[i]);
      if (zero <= 0.0)
        { errorExit("zero value must be strictly positive"); }
      i++;
      continue;
      }

    /* Options with two parameters ... none */

    /* If we fall through to here, the argument was unrecognised and it's an
    error.  At some point in the future, unrecognised arguments might be, eg,
    treated as filenames (if they don't start with "-"), and processed here. */

    fprintf(stderr, "ERROR: unrecognised argument, \"%s\"\n", argv[i]);
    exit(EXIT_FAILURE);
    }
  }

/* Check the parameters & do some setup/initialisation of IG library. */

void chkSetup(void)
  {
  /* Seed the random number generator and set all the igfp stuff, except for the
  function. */

  igSeedRand(seed);

  igDim = dim;
  igTol = eps;
  igZero = zero;
  igPLimit = piv;

  if      (lhpivot == 2) { igLHPivot = igLHbest; }
  else if (lhpivot == 1) { igLHPivot = igLHnorm; }
  else                   { igLHPivot = igLHlast; }

  if      (norm == -1) { igNorm = *igLinfNorm; }
  else if (norm == 1)  { igNorm = *igL1Norm; }
  else                 { igNorm = *igL2Norm; }
  }

/*******************************************************************************
The entry point and top-level wrapper for the whole shebang ...
*******************************************************************************/

int main(int argc, char *argv[])
  {
  int a, i, j, k;
  double n, *x, *y, *tp;
  igList *igl = NULL;
  igFPSTAT igfps;

  initParam();
  procArg(argc, argv);
  chkSetup();

  printf("\n");
  printBanner();
  printf("\n");
  printParam();

  /* Setup the function to use for all the attempts. */

  f1Init();
  printf("\n");
  f1Print();
  igFunction = *f1Apply;

  for (a = 1; a <= att; a++)
    {
    printf("\n");
    if (a == 1) { printf("Barycentric attempt ...\n"); }
    else        { printf("Attempt #%d ...\n", a); }

    callCnt = 0;

    /* Generate a starting X/Y/norm triple.  The first time through we use the
    barycenter, then we use 'random' stochastic vectors. */

    if (a == 1)
      { x = igGetCVect(dim, 1.0/(double)dim); }
    else
      {
      x = igGetVect(dim);
      i = 0;
      for (j = 0; j < dim; j++) 
        { 
        k = 1 + rand()%(32+fmode+dim);
        x[j] = (double)k;
        i += k;
        }
      for (j = 0; j < dim; j++) { x[j] /= (double)i; }
      }

    y = igGetVect(dim);

    if (f1Apply(x, y))
      {
      n = (*igNorm)(igDim, x, y);
      if (n < igTol)
        {
        printf("  FP, initial phase\n");
        printf("    norm X-Y = %0.15g\n", n);
        printf("    E-value = %0.15g\n", lambda);
        printf("    E-vect =");
        for (i = 0; i < dim; i++) { printf(" %0.15g", x[i]); }
        printf("\n");

        free(x);
        free(y);
        continue;
        }
      }
    else
      {
      printf("  initial phase failed\n");
      free(x);
      free(y);
      continue;
      }

    /* If we fall through to here, we successfully generated a non-FP X/Y/norm
    triple.  Create a new igList and add the starting triple to it. */

    if (igl != NULL) { igFreeList(igl); }
    igl = igNewList(igDim);

    igAppend(igl, x, y, n);

    /* Now run the imitate or iterate FP finder. */

    if (meth == 1)     /* use imitation games */
      {
      /* At this point the list contains precisely 1 point, so the first IG game
      will have size 1x1.  The lim parameter is at least 2 and, to ensure that 
      the largest IG has size at most lim, we pass lim as the "cnt" argument to
      igImitate(). */

      igfps = igImitate(igl, lim, 0U);
      if (igfps == igFP)
        {
        printf("  FP, via imitation, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
        printf("    norm X-Y = %0.15g\n", igl->dn[igl->size-1]);
        printf("    E-value = %0.15g\n", lambda);
        printf("    E-vect =");
        for (i = 0; i < dim; i++) 
          { printf(" %0.15g", igl->dx[igl->size-1][i]); }
        printf("\n");

        continue;
        }
      else if (igfps == igOK)
        { ; }
      else
        {
        printf("  imitation phase failed, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
        continue;
        }
      }
    else               /* use iteration */
      {
      igfps = igIterate(igl, lim);
      if (igfps == igFP)
        {
        printf("  FP, via iteration, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
        printf("    norm X-Y = %0.15g\n", igl->dn[igl->size-1]);
        printf("    E-value = %0.15g\n", lambda);
        printf("    E-vect =");
        for (i = 0; i < dim; i++)
          { printf(" %0.15g", igl->dx[igl->size-1][i]); }
        printf("\n");

        continue;
        }
      else if (igfps == igOK)
        { ; }
      else
        {
        printf("  iteration phase failed, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
        continue;
        }
      }

    /* If we fall through to here there haven't been any problems but we haven't
    found a within-tolerance FP.  We have however appended one or more X/Y/norm
    tuples to the list, so we dump the best approx'n found, and then see what we
    can do about polishing it.  NB: We cannot use the lambda value, since that's
    for the last function call which, in general, will not be for the list entry
    chosen.*/

    k = igl->size - 1;     /* last index for current igList */

    n = igl->dn[0];
    j = 0;
    for (i = 1; i <= k; i++)
      { if (igl->dn[i] < n) { n = igl->dn[i];  j = i; } }

    printf("  best approximate FP, |list| = %d, callCnt = %d\n", k+1, callCnt);
    printf("    norm X-Y = %0.15g\n", n);
    printf("    X[%d] =", j);
    for (i = 0; i < dim; i++) { printf(" %0.15g", igl->dx[j][i]); }
    printf("\n");

    /* Polish uses the last entry on the list, so we swap the least & last. */

    if (j < k)
      {
      tp = igl->dx[j];  igl->dx[j] = igl->dx[k];  igl->dx[k] = tp;
      tp = igl->dy[j];  igl->dy[j] = igl->dy[k];  igl->dy[k] = tp;
      n  = igl->dn[j];  igl->dn[j] = igl->dn[k];  igl->dn[k] = n;
      }

    igfps = igPolish(igl, gsliter);
    if (igfps == igFP)
      {
      printf("  FP, via polishing, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
      printf("    norm X-Y = %0.15g\n", igl->dn[igl->size-1]);
      printf("    E-value = %0.15g\n", lambda);
      printf("    E-vect =");
      for (i = 0; i < dim; i++)
        { printf(" %0.15g", igl->dx[igl->size-1][i]); }
      printf("\n");
      }
    else if (igfps == igOK)
      {
      printf("  best polished FP, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
      printf("    norm X-Y = %0.15g\n", igl->dn[igl->size-1]);
      printf("    E-value ~= %0.15g\n", lambda);
      printf("    E-vect ~=");
      for (i = 0; i < dim; i++)
        { printf(" %0.15g", igl->dx[igl->size-1][i]); }
      printf("\n");
      }
    else
      {
      printf("  polishing phase failed, |list| = %d, callCnt = %d\n", 
                                                            igl->size, callCnt);
      continue;
      }
    }

  f1Free();
  if (igl != NULL) { igFreeList(igl); }

  printf("\n");
  exit(EXIT_SUCCESS);
  }


