
/*******************************************************************************

    example3.c
    25 Apr 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

This is an example programme using a class of functions based on Example 4.6 of 
"From Imitation Games to Kakutani" by A. McLennan and R. Tourkey (20 Mar 06).  A
random example is generated, and the igig wrapper is used to solve it.  The 
functions are high dimensional, but are 'well-behaved' and 'easy' to solve.  If
the defaults are used then an interesting change in the 'average' difficulty can
be seen (set the debug flag) as the dimension is increased above ~400.

*******************************************************************************/

#include "igig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsl/gsl_version.h>

/*******************************************************************************
All of the parameters for this example programme are accessible anywhere within 
this file, and they are passed as parameters to the IG library as required.  Any
features of the library not used, or not settable by the user, as set to default
values (either explicitly or implicitly).
*******************************************************************************/

unsigned int debug, seed;
int dim, fmode, norm;
double xmode, eps, zero;
int lim, piv;
int att, lhpivot, gsliter;
int guess, loop;

/*******************************************************************************
The dimension of the space is "din' and the generation of M is set by "min:
  1    M = 3I
  2    use absolute values of random values from normal distribution, N(0,1)
  3    use random values from normal distribution, N(0,1/13)
*******************************************************************************/

int f3dim;             /* the dimension of the function space */
int f3mode;            /* the mode (1, 2 or 3) for generating M */
double *f3z;           /* the Z point */
double **f3m;          /* the M matrix */
double *f3b;           /* a working vector */
int f3call;            /* number of function calls */

void f3Init(int din, int min)
  {
  int i,j;
  double tmp = 1.0/13.0;

  f3dim = din;
  f3mode = min;

  f3z = igGetNVect(f3dim, 0.0, 1.0);

  if ((f3m = (double**)malloc(f3dim * sizeof(double*))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in f3Init()"); }

  switch (f3mode)
    {
    case 1:
      for (i = 0; i < f3dim; i++)
        {
        f3m[i] = igGetCVect(f3dim, 0.0);
        f3m[i][i] = 3.0;
        }
      break;

    case 2:
      for (i = 0; i < f3dim; i++)
        { 
        f3m[i] = igGetNVect(f3dim, 0.0, 1.0);
        for (j = 0; j < f3dim; j++) { f3m[i][j] = fabs(f3m[i][j]); }
        }
      break;

    case 3:
      for (i = 0; i < f3dim; i++)
        { f3m[i] = igGetNVect(f3dim, 0.0, tmp); }
      break;

    default:
      igAbend(__FILE__, __LINE__, "bad M mode in f3Init()");
    }

  f3b = igGetVect(f3dim);
  f3call = 0;
  }

void f3Free(void)
  {
  if (f3z != NULL) { free(f3z); }
  if (f3m != NULL)
    {
    for (int i = 0; i < f3dim; i++) { free(f3m[i]); }
    free(f3m);
    }
  if (f3b != NULL) { free(f3b); }
  }

/* f(X) = Z - arctan(M * (X-Z)^3) */

igBOOL f3Apply(double *x, double *y)
  {
  double c;
  int i, j;

  for (i = 0; i < f3dim; i++) { y[i] = f3z[i]; }

  for (i = 0; i < f3dim; i++) { c = x[i] - f3z[i];  f3b[i] = c*c*c; }

  for (i = 0; i < f3dim; i++)
    {
    c = 0.0;
    for (j = 0; j < f3dim; j++) { c += f3m[i][j] * f3b[j]; }
    y[i] -= atan(c);
    }

  f3call++;
  return igTRUE;
  }

/*******************************************************************************
These are the utility programmes we need ...
*******************************************************************************/

/* This is for user (ie, parameter) errors, not IG library errors. */

void errorExit(char *s)
  {
  fprintf(stderr, "ERROR: %s\n", s);
  exit(EXIT_FAILURE);
  }

/* Print the opening banner message. */

void printBanner(void)
  {
  printf("Find a fixed point, via imitation games and root polishing\n");
  printf("%s is Copyright (C) 2011, The Gambit Project\n", IG);
  printf("Incorporating GSL version %s\n", GSL_VERSION);
  printf("This is free software, distributed under the GNU GPL\n");
  }

/* Print the help message. */

void printHelp(char *s)
  {
  printf("\n");
  printf("Usage: %s [OPTIONS]\n", s);
  printf("Options:\n");
  printf("  --attempt A    Make A (> 0) attempts to find a root\n");
  printf("  -(-d)ebug A    Set the debug output flags to A (>= 0)\n");
  printf("  --dim A        Set the dimension of the problem to A (> 0)\n");
  printf("  --eps A        Set the norm tolerance to A (> 0.0)\n");
  printf("  --fmode A      Set the function mode to A (1, 2 or 3)\n");
  printf("  --guess A      Make A (> 0) guesses for initial X\n");
  printf("  -(-h)elp       Print this help screen and then terminate\n");
  printf("  --iter A       Set the GSL iteration limit to A (> 0)\n");
  printf("  --lh A         Set Lemke-Howson pivot mode to A (0, 1 or 2)\n");
  printf("  --limit A      Set maximum imitation game size to A (> 1)\n");
  printf("  --loop A       Make up to A (> 0) inner loops\n");
  printf("  --norm A       Set the norm to use to A (-1, 1 or 2)\n");
  printf("  --pivot A      Set maximum pivots per Lemke-Howson to A (> 1)\n");
  printf("  --seed A       Seed the random number generator with A (> 0)\n");
  printf("  --xmode A      Set the initial X mode to A (>= 0.25)\n");
  printf("  --zero A       Set the floating-point tolerance to A (> 0.0)\n");
  printf("\n");
  }

/* Print the IG settings & any other parameters. */

void printParam(void)
  {
  printf("dim = %d", igDim);
  if (fmode == 1)
    { printf("  fmode = 3I_d"); }
  else if (fmode == 2)
    { printf("  fmode = |N(0,1)|"); }
  else
    { printf("  fmode = N(0,1/13)"); }
  printf("  xmode = [%0.6g,%0.6g]  debug = %u\n", igigDmin, igigDmax, debug);

  printf("seed = %u", igRandomSeed);
  if (igLHPivot == igLHbest) 
    { printf("  lhpivot = best"); }
  else if (igLHPivot == igLHnorm) 
    { printf("  lhpivot = norm"); }
  else  
    { printf("  lhpivot = last"); }
  if (igNorm == *igLinfNorm)
    { printf("  norm = Linf\n"); }
  else if (igNorm == *igL1Norm)
    { printf("  norm = L1\n"); }
  else
    { printf("  norm = L2\n"); }

  printf("limit = %d  pivot = %d  eps = %0.15g  zero = %0.15g\n", 
                                           igigLimit1, igPLimit, igTol, igZero);

  printf("attempt = %d  guess = %d  iter = %d  loop = %d\n", 
                             igigAttempts, igigGuesses, igigLimit2, igigLimit3);
  }

/* Set all the parameters to their initial default values. */

void initParam(void)
  {
  att = 5;             /* make up to five attempts */
  debug = 0U;          /* no debugging output */
  dim = 250;           /* dimension of the problem space */
  eps = 1.0E-9;        /* epsilon for testing norm */
  fmode = 3;           /* function mode; M is drawn from N(0,1/13) */
  gsliter = 50;        /* limit on GSL polishing steps */
  guess = 250;         /* init X is best of 250 guesses */
  lhpivot = 0;         /* use last X as initial pivot */
  lim = 50;            /* max imitation game size */
  loop = 8;            /* make up to eight inner loops */
  norm = -1;           /* use Linfinity norm */
  piv = 75;            /* max pivots per Lemke-Howson call */
  seed = 0U;           /* generate a (random) seed internally */
  xmode = 1.0;         /* initial X mode; drawn uniformly from [-1.0,+1.0] */
  zero = 4.0E-15;      /* floating-pt zero tolerance */
  }

/* Process all the arguments and set the parameters as directed. */

void procArg(int argc, char *argv[])
  {
  int i = 1;

  while (i < argc)
    {
    /* These options will terminate the programme */

    if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0) 
      { printHelp(argv[0]);  exit(0); }

    /* Options with no parameters */

    /* Options with one parameter */

    if (strcmp(argv[i], "--attempt") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--attempt\""); }
      att = atoi(argv[i]);
      if (att < 1)
        { errorExit("number of attempts must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--debug") == 0 || strcmp(argv[i], "-d") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--debug/-d\""); }
      debug = (unsigned int)strtoul(argv[i], NULL, 10);
      if (debug < 0)         /* can't happen! */
        { errorExit("debug setting must be 0 or greater"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--dim") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--dim\""); }
      dim = atoi(argv[i]);
      if (dim < 1)
        { errorExit("space must have dimension at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--eps") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--eps\""); }
      eps = atof(argv[i]);
      if (eps <= 0.0)
        { errorExit("epsilon value must be strictly positive"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--fmode") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--fmode\""); }
      fmode = atoi(argv[i]);
      if (fmode < 1 || fmode > 3)
        { errorExit("function mode must be 1, 2 or 3"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--guess") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--guess\""); }
      guess = atoi(argv[i]);
      if (guess < 1)
        { errorExit("number of initial guesses must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--iter") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--iter\""); }
      gsliter = atoi(argv[i]);
      if (gsliter < 1)
        { errorExit("GSL iteration limit must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--lh") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--lh\""); }
      lhpivot = atoi(argv[i]);
      if (lhpivot < 0 || lhpivot > 2)
        { errorExit("LH pivoting choice method must be 0, 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--limit") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--limit\""); }
      lim = atoi(argv[i]);
      if (lim < 2)
        { errorExit("game size limit must be at least 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--loop") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--loop\""); }
      loop = atoi(argv[i]);
      if (loop < 1)
        { errorExit("inner loop limit must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--norm") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--norm\""); }
      norm = atoi(argv[i]);
      if (norm != -1 && norm != 1 && norm != 2)
        { errorExit("norm must be -1 (inf), 1 or 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--pivot") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--pivot\""); }
      piv = atoi(argv[i]);
      if (piv < 2)
        { errorExit("pivot limit must be at least 2"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--seed") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--seed\""); }
      seed = (unsigned int)strtoul(argv[i], NULL, 10);
      if (seed < 1)
        { errorExit("seed must be at least 1"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--xmode") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--xmode\""); }
      xmode = atof(argv[i]);
      if (xmode < 1.0/4.0)
        { errorExit("initial X mode must be at least 1/4"); }
      i++;
      continue;
      }

    if (strcmp(argv[i], "--zero") == 0)
      {
      i++;
      if (!(i < argc))
        { errorExit("missing argument for \"--zero\""); }
      zero = atof(argv[i]);
      if (zero <= 0.0)
        { errorExit("zero value must be strictly positive"); }
      i++;
      continue;
      }

    /* Options with two parameters */

    /* If we fall through to here, the argument was unrecognised and it's an
    error.  At some point in the future, unrecognised arguments might be, eg,
    treated as filenames (if they don't start with "-"), and processed here. */

    fprintf(stderr, "ERROR: unrecognised argument, \"%s\"\n", argv[i]);
    exit(EXIT_FAILURE);
    }
  }

/* Check the parameters and do the setup/initialisation of the igig library and
the function.  Note that igigSetup() is not used, since all the parameters are
set by this programmes arguments. */

void chkSetup(void)
  {
  igSeedRand(seed);

  igTol = eps;
  igZero = zero;
  igPLimit = piv;

  if      (lhpivot == 2) { igLHPivot = igLHbest; }
  else if (lhpivot == 1) { igLHPivot = igLHnorm; }
  else                   { igLHPivot = igLHlast; }

  if      (norm == -1) { igNorm = *igLinfNorm; }
  else if (norm == 1)  { igNorm = *igL1Norm; }
  else                 { igNorm = *igL2Norm; }

  igigAttempts = att;
  igigGuesses = guess;
  igigLimit1 = lim;
  igigLimit2 = gsliter;
  igigLimit3 = loop; 

  f3Init(dim, fmode);
  igDim = dim;
  igFunction = *f3Apply;

  igigDmin = -xmode;
  igigDmax = +xmode;
  }

/*******************************************************************************
The entry point and top-level wrapper for the whole shebang ...
*******************************************************************************/

int main(int argc, char *argv[])
  {
  igList *igl;
  igFPSTAT igfps;

  initParam();
  procArg(argc, argv);
  chkSetup();

  printf("\n");
  printBanner();
  printf("\n");
  printParam();

  igl = igNewList(igDim);            /* create the results list */
  igfps = igigFindFP(igl, debug);    /* try to find an FP */

  printf("\n");
  if (igfps == igFail)
    { printf("Unable to find any (approx) fixed point\n"); }
  else
    {
    if (igfps == igFP)
      { printf("Found a fixed point, norm = %0.15g\n", igl->dn[igl->size-1]); }
    else
      { printf("Best approx'n found, norm = %0.15g\n", igl->dn[igl->size-1]); }
    }
  printf("  fn calls = %d\n", f3call);

  free(igl);
  f3Free();

  printf("\n");
  exit(EXIT_SUCCESS);
  }


