
/*******************************************************************************

    ig.c
    1 Jun 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "ig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <sys/types.h>     /* for getpid() */
#include <unistd.h>

/* Error handling is very simple.  We just print out where we were and what went
wrong and then exit. */

void igAbend(char *file, int line, char *msg)
  {
  fprintf(stderr, "ERROR at line %d of file %s\n", line, file);
  fprintf(stderr, "  %s\n", msg);
  exit(EXIT_FAILURE);
  }

unsigned int igRandomSeed;

void igSeedRand(unsigned int s)
  {
  while (s == 0U)  
    { s = (unsigned int)time(NULL) + (unsigned int)clock() 
                                                     + (unsigned int)getpid(); }
  srand(s);
  igRandomSeed = s;
  }

/*******************************************************************************
This is from "Numerical Recipes in C", 2nd Edition, pp 288/290.  Note that only 
one of the possible two values is used.  TBA: Should we default the mean and std
dev'n to 0.0 and 1.0?
*******************************************************************************/

double igNormal(double mean, double sd)
  {
  double v1, v2, rSq;
  igBOOL done = igFALSE;

  while (!done)
    {
    v1 = 2.0 * ((double)rand()/(double)RAND_MAX) - 1.0;
    v2 = 2.0 * ((double)rand()/(double)RAND_MAX) - 1.0;
    rSq = v1*v1 + v2*v2;
    if (rSq < 1.0 && rSq != 0.0)    /* inside unit circle & non-zero */ 
      { done = igTRUE; }     
    }

  return mean + sd * v1 * sqrt(-2.0 * log(rSq) / rSq);
  }

double* igGetNVect(int dim, double mean, double sd)
  {
  double *p;
  int i;

  if ((p = (double*)malloc(dim * sizeof(double))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igGetNVect()"); }

  for (i = 0; i < dim; i++)
    { p[i] = igNormal(mean, sd); }

  return p;
  }

/* TBA: Change interval to open, by disallowing the 0 and RAND_MAX values from
rand()? */

double* igGetUVect(int dim, double min, double max)
  {
  double *p;
  int i;

  if ((p = (double*)malloc(dim * sizeof(double))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igGetUVect()"); }

  for (i = 0; i < dim; i++)
    { p[i] = min + (max-min) * ((double)rand()/(double)RAND_MAX); }

  return p;
  }

double* igGetCVect(int dim, double c)
  {
  double *p;
  int i;

  if ((p = (double*)malloc(dim * sizeof(double))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igGetCVect()"); }

  for (i = 0; i < dim; i++)
    { p[i] = c; }

  return p;
  }

double* igGetVect(int dim)
  {
  double *p;
  if ((p = (double*)malloc(dim * sizeof(double))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igGetVect()"); }
  return p;
  }

double* igReadVect(int dim)
  {
  int i;
  double *p, d;

  if ((p = (double*)malloc(dim * sizeof(double))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igReadVect()"); }

  for (i = 0; i < dim; i++)
    {
    if (scanf(" %lf", &d) != 1)
      { igAbend(__FILE__, __LINE__, "can't read double in igReadVect()"); }
    p[i] = d;
    }

  return p;
  }

double* igCopyVect(int dim, double *v)
  {
  int i;
  double *p;

  if ((p = (double*)malloc(dim * sizeof(double))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igCopyVect()"); }

  for (i = 0; i < dim; i++) { p[i] = v[i]; }

  return p;
  }

/* NB: There is no checking for NAN/INF in these. */

double igL1Norm(int dim, double *v1, double *v2)
  {
  int i;
  double n = 0.0;
  for (i = 0; i < dim; i++)
    { n += fabs(v1[i] - v2[i]); }
  return n;
  }

double igL2Norm(int dim, double *v1, double *v2)
  {
  int i;
  double n = 0.0;
  for (i = 0; i < dim; i++)
    { n += (v1[i] - v2[i]) * (v1[i] - v2[i]); }
  return sqrt(n);
  }

double igLinfNorm(int dim, double *v1, double *v2)
  {
  int i;
  double m, n = 0.0;
  for (i = 0; i < dim; i++)
    { if ((m = fabs(v1[i] - v2[i])) > n) { n = m; } }
  return n;
  }

/*******************************************************************************
Prints the elements of the vector 4-to-a-row with full 15-digit precision.  The
final line has a trailing newline on it.  If the string "s" is NULL, only the 
vector is printed.  If not, a heading line containing the string is printed.
NB: The magic numbers (4 & 15) are embedded in the code.
*******************************************************************************/

void igPrettyP(int dim, double *v, char *s)
  {
  int i;
  if (s != NULL) { printf("%s :-\n", s); }
  for (i = 0; i < dim; i++)
    {
    if (i%4 == 0) { printf(" "); }
    printf(" %0.15g", v[i]);
    if (i%4 == 3) { printf("\n"); }
    }
  if (dim%4 != 0) { printf("\n"); }
  }

/* Returns true if all the values are finite (ie, not INF or NAN). */

igBOOL igVectFinite(int dim, double *v)
  {
  int i;
  for (i = 0; i < dim; i++)
    { if (!isfinite(v[i])) { return igFALSE; } }
  return igTRUE;
  }

/* Checks each coordinate (assumed to be finite) against the interval [min,max].
Values outside this range are clipped.  Any clipping means a true return. */

igBOOL igClipVect(int dim, double *v, double min, double max)
  {
  int i;
  igBOOL b = igFALSE;
  for (i = 0; i < dim; i++)
    {
    if      (v[i] < min) { v[i] = min;  b = igTRUE; }
    else if (v[i] > max) { v[i] = max;  b = igTRUE; }
    }
  return b;
  }

/* Returns true if one or more of the coordinates is within a tolerance of "tol"
of the upper or lower boundary. */

igBOOL igOnBound(int dim, double *v, double min, double max, double tol)
  {
  int i;
  for (i = 0; i < dim; i++)
    {
    if (fabs(v[i]-min) < tol) { return igTRUE; }
    if (fabs(v[i]-max) < tol) { return igTRUE; }
    }
  return igFALSE;
  }


