
/*******************************************************************************

    igfp.c
    3 Jun 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "igfp.h"
#include "igtableau.h"

#include <math.h>
#include <stdio.h>     /* Only needed if logging/debug output is uncommented */
#include <stdlib.h>

int igDim;
double igTol;
double igZero;
int igPLimit;

igLHPIV igLHPivot;

igFUN igFunction;
igNORM igNorm;

/*******************************************************************************
Tries to find a 'good' initial X by repeatedly generating & testing random X, 
drawn uniformly from the nominated region, and keeping track of the one with the
closest Y value.  The best X/Y/norm is appended to the igList, and igFP/igOK (as
appropriate) is returned.  If no valid X/Y/norm triple is generated, within the 
given limit on the number of attempts, the call fails.  The loop is short-
circuited if an FP is found.  NB: rmin < rmax is enforced.
*******************************************************************************/

igFPSTAT igGuess(igList *igl, int cnt, double rmin, double rmax)
  {
  int i;
  double *x0 = NULL, *y0 = NULL, *x1, *y1 = NULL;
  double n0 = -2.0, n1;

  if (cnt < 1 || rmin >= rmax) { return igFail; }

  for (i = 1; i <= cnt; i++)
    {
    x1 = igGetUVect(igDim, rmin, rmax);
    if (y1 == NULL) { y1 = igGetVect(igDim); }

    if (!(*igFunction)(x1, y1))   { free(x1);  continue; }
    if (!igVectFinite(igDim, y1)) { free(x1);  continue; }

    n1 = (*igNorm)(igDim, x1, y1);
    if (!isfinite(n1)) { free(x1);  continue; }

    if (n0 < -1.0)
      { x0 = x1;  y0 = y1;  n0 = n1;  y1 = NULL; }
    else
      {
      if (n1 < n0)
        { free(x0);  free(y0);  x0 = x1;  y0 = y1;  n0 = n1;  y1 = NULL; }
      else
        { free(x1); }
      }

    if (n0 < igTol) { break; } 
    }

  if (y1 != NULL) { free(y1); }

  if (n0 < -1.0) { return igFail; }    /* no valid triple generated */

  igAppend(igl, x0, y0, n0);

  if (n0 < igTol) { return igFP; }     /* within tolerance fixed-point */
  return igOK;                         /* best approximation */
  }

/*******************************************************************************
Given a non-empty igList, this iterates the function starting at the last Y 
point in the list.  Up to cnt iterations are performed, and we stop if get a
within tolerance FP, hit the limit, or have a problem.  The final X/Y/norm 
triple generated (if any) is appended to the list.  If no valid X/Y/norm tuple 
is appended to the list, the call fails.
TBA:  Store all the X/Y/norm tuples generated?
*******************************************************************************/

igFPSTAT igIterate(igList *igl, int cnt)
  {
  int iter;                    /* current iteration */
  double *px;                  /* previous X */
  double *py;                  /* previous Y */
  double pn;                   /* previous norm */
  double *cx;                  /* current X */
  double *cy;                  /* current Y */
  double cn;                   /* current norm */

  if (igl->size < 1 || cnt < 1) { return igFail; }

  px = py = NULL;  pn = -2.0;

  for (iter = 1; iter <= cnt; iter++)
    {
    if (iter == 1)
      {
      /* The first time through is special, since we have not yet generated any
      triples and the X to use is the final Y in the list.  So we need to take
      care, and set things up for the subsequent passes (if any).  On error we
      free any newly allocated memory & return.  If we hit an FP we append the 
      result to the list & return. */

      cx = igl->dy[igl->size-1];     /* the last Y value */
      cy = igGetVect(igDim);         /* space for the result */

      if (!(*igFunction)(cx, cy))   { free(cy);  return igFail; }
      if (!igVectFinite(igDim, cy)) { free(cy);  return igFail; }

      cn = (*igNorm)(igDim, cx, cy);
      if (!isfinite(cn)) { free(cy);  return igFail; }

      px = igCopyVect(igDim, cx);
      py = cy;
      pn = cn;

      if (pn < igTol)
        {
        igAppend(igl, px, py, pn);
        return igFP;
        }
      }
    else
      {
      /* For the 2nd & subsequent iterations, it is guaranteed that a valid
      previous X/Y/norm triple exists.  If there's an error, this has to be
      appended to the list and igOK returned. */

      cx = py;
      cy = igGetVect(igDim);

      if (!(*igFunction)(cx, cy))   
        { igAppend(igl, px, py, pn);  free(cy);  return igOK; }
      if (!igVectFinite(igDim, cy)) 
        { igAppend(igl, px, py, pn);  free(cy);  return igOK; }

      cn = (*igNorm)(igDim, cx, cy);
      if (!isfinite(cn)) 
        { igAppend(igl, px, py, pn);  free(cy);  return igOK; }

      free(px);
      px = py;
      py = cy;
      pn = cn;

      if (pn < igTol)
        {
        igAppend(igl, px, py, pn);
        return igFP;
        }
      }
    }

  /* We fall through to here if there were no errors and we didn't find an FP.
  We simply append the last triple found to the list and return. */

  igAppend(igl, px, py, pn);
  return igOK;
  }

/*******************************************************************************
Given a non-empty igList, this builds and solves the imitation game (aka, IG) 
for the list and generates an X/Y/norm triple.  This is appended to the list and
the process repeated.  We stop after cnt games have been built & solved, or we 
find an FP, or there's an error.  Note that all the valid triples generated are 
appended to the list, since they're needed to build the IGs.  The FP (if found) 
will be the last item on the list.  If no items were added to the list, the call
fails.  

Note that if the list contains a single entry, then building and solving the 
initial IG is equivalent to function iteration.  This is not necessarily true 
for subsequent IGs.  If the initial list size is L and the process doesn't stop
until all cnt IGs are built & solved, then the largest IG has size L+cnt-1.

The debug flag array can be used to dump extra information:
  debug & 0x1    pretty print the Y prob'y vector used to generate the next X
          0x2    trace the execution of an igLHPivot == igLHbest run

*******************************************************************************/

igFPSTAT igImitate(igList *igl, int cnt, unsigned int debug)
  {
  igFPSTAT fpstat = igFail;          /* default return status */
  int stat = -1;                     /* unchanged on normal loop termination */
  igMatrix *igm = NULL;              /* current IG norm matrix */
  igTableau *igt = NULL;             /* current tableau */
  double *x = NULL, *y = NULL;       /* current X & Y points */
  double *py = NULL;                 /* current IG solution Y prob'y vector */
  double *tp = NULL, *tx = NULL;     /* temp prob & X for igLHbest mode */
  int iter, size, cindex, i, j, k;
  double n, p, *t, tn;

  /* Check initial conditions, and build the starting -norm^2 matrix.  Note that
  this matrix uses the L2-norm, not the FP-tolerance norm. */

  size = igl->size;
  if (size < 1 || cnt < 1) { return fpstat; }

  igm = igNewMatrix();
  for (i = 0; i < size; i++) { igResize(igm); }

  for (i = 0; i < size; i++)
    {
    for (j = 0; j < size; j++)
      {
      n = igL2Norm(igDim, igl->dx[i], igl->dy[j]);
      n = -(n*n);
      if (!isfinite(n))
        { igFreeMatrix(igm);  return fpstat; }
      igSetIJ(igm, i, j, n);
      }
    }

  cindex = size-1;     /* initial index of last data in igm/igl */

  for (iter = 1; iter <= cnt; iter++)
    {
    /* Note that there's no setup to do for the 1st iteration, since the matrix 
    & the data index have already been initialised.  On subsequent iterations, 
    the X/Y values from the previous iteration need to be added to the matrix.  
    They've already been added to the list (along with the norm). */

    if (iter > 1)
      {
      cindex++;
      igResize(igm);

      for (i = 0; i < cindex; i++)     /* new column, ex last row */
        {
        n = igL2Norm(igDim, igl->dx[i], igl->dy[cindex]);
        n = -(n*n);
        if (!isfinite(n)) { stat = 1;  break; }
        igSetIJ(igm, i, cindex, n);
        }
      if (stat != -1) { break; }

      for (i = 0; i <= cindex; i++)    /* new row */
        {
        n = igL2Norm(igDim, igl->dx[cindex], igl->dy[i]);
        n = -(n*n);
        if (!isfinite(n)) { stat = 1;  break; }
        igSetIJ(igm, cindex, i, n);
        }
      if (stat != -1) { break; }
      }

    if (igt != NULL) { igFreeTableau(igt); }
    igt = igNewTableau(igm, igZero);

    /* In principle, the initial pivot for Lemke-Howson can be any of the X or Y
    points.  The default is igLHlast, which uses the last (ie, most recent) X 
    point on the list.  Note that LH pivots are numbered starting at 1 and
    points are indexed starting at 0, hence the "+1". */

    if (igLHPivot == igLHnorm)         /* use X with smallest X-Y norm */
      {
      n = igl->dn[0];
      j = 0;
      for (i = 1; i <= cindex; i++)
        { if (igl->dn[i] < n) { n = igl->dn[i];  j = i; } }

      if (igLemHow(igt, j+1, igPLimit) != igNE) 
        { stat = 1;  break; }
      }
    else if (igLHPivot == igLHbest)    /* use X yielding best new X ... */
      {                                /* ... try the 1st X             */
      if (igLemHow(igt, 1, igPLimit) != igNE) 
        { stat = 1;  break; }
      }
    else                               /* default, use most recent X */
      { 
      if (igLemHow(igt, cindex+1, igPLimit) != igNE) 
        { stat = 1;  break; }
      }

    if (py != NULL) { free(py); }
    py = igGetYProb(igt);
    if (py == NULL) { stat = 1;  break; }

    /* The next X point is the probability-weighted sum of the Yi. */

    if (x != NULL) { free(x); }
    x = igGetCVect(igDim, 0.0);

    for (i = 0; i <= cindex; i++)
      {
      if ((p = py[i]) > 0.0)
        { 
        t = igl->dy[i];
        for (j = 0; j < igDim; j++) { x[j] += p*t[j]; }
        }
      }

    /* At this point we have a NE, its Y strategy vector (py) and matching X
    point (x).  Unless we're in igLHbest mode and the game has size at least 
    two, we're ready to proceed. If not, we need to try all other pivot choices 
    first, to see if we can generate a 'better' Y strategy vector (tp) to use as
    the basis for generating the X point to 'use' as the pivot.  The debug & 0x2
    code can be used to monitor the Y strategy vectors generated by the 
    Lemke-Howson process as we run through all pivots (X strategies). */

    if (igLHPivot == igLHbest && cindex > 0)
      {
      /* We need to repeat the previous block of code for all the other X as the
      initial pivot, and track which one yields a new X closest to the X from
      the previous iteration. */

      /* First, record how close our 1st choice of pivot got. */

      n = (*igNorm)(igDim, x, igl->dx[cindex]);
      if (!isfinite(n)) { stat = 1;  break; }

      if (debug & 0x2)
        {
        igPrettyP(cindex+1, py, "prob, 1st strategy");
        printf("norm [1] = %0.15g\n", n);
        }

      /* Now loop over all other choices of pivot. */

      for (k = 1; k <= cindex; k++)
        {
        if (igt != NULL) { igFreeTableau(igt); }
        igt = igNewTableau(igm, igZero);

        if (igLemHow(igt, k+1, igPLimit) != igNE) 
          { stat = 1;  break; }

        if (tp != NULL) { free(tp); }
        tp = igGetYProb(igt);
        if (tp == NULL) { stat = 1;  break; }

        if (tx != NULL) { free(tx); }
        tx = igGetCVect(igDim, 0.0);

        for (i = 0; i <= cindex; i++)
          {
          if ((p = tp[i]) > 0.0)
            { 
            t = igl->dy[i];
            for (j = 0; j < igDim; j++) { tx[j] += p*t[j]; }
            }
          }

        tn = (*igNorm)(igDim, tx, igl->dx[cindex]);
        if (!isfinite(tn)) { stat = 1;  break; }

        if (debug & 0x2)
          {
          igPrettyP(cindex+1, tp, "prob, next strategy");
          printf("norm [%d] = %0.15g\n", k+1, tn);
          }

        if (tn < n)
          {
          free(py);  py = tp;  tp = NULL;
          free(x);    x = tx;  tx = NULL;
          n = tn;

          if (debug & 0x2)
            {
            igPrettyP(cindex+1, py, "prob, updated best");
            printf("norm, updated best = %0.15g\n", n);
            }
          }
        }

      /* Finally, detect if there were any problems. */

      if (stat != -1) { break; }
      }

    /* We now have the last/norm/best X point generated from the IG. */

    if (debug & 0x1)
      { igPrettyP(cindex+1, py, "Y prob"); }

    if (y == NULL) { y = igGetVect(igDim); }

    if (!(*igFunction)(x, y))    { stat = 1;  break; }
    if (!igVectFinite(igDim, y)) { stat = 1;  break; }

    n = (*igNorm)(igDim, x, y);
    if (!isfinite(n)) { stat = 1;  break; }

    /* If we fall through to here, we've successfully generated a new X/Y/norm
    tuple from the IG.  We need to append this to the list, clear some pointers 
    (without deallocating) ready for the next iteration, and test for an FP. */

    igAppend(igl, x, y, n);
    x = NULL;
    y = NULL;

    if (n < igTol) { stat = 0;  break; }
    }

  /* Tidy up and return result, based on how we exited the loop. */

  if (igm != NULL) { igFreeMatrix(igm); }
  if (igt != NULL) { igFreeTableau(igt); }

  if (x != NULL) { free(x); }
  if (y != NULL) { free(y); }

  if (py != NULL) { free(py); }

  if (tp != NULL) { free(tp); }
  if (tx != NULL) { free(tx); }

  switch (stat)
    {
    case -1:     /* normal loop termination, with no FP */
      fpstat = igOK;
      break;

    case 0:      /* FP induced loop termination */
      fpstat = igFP;
      break;

    case 1:      /* error induced loop termination */
      if (igl->size > size) { fpstat = igOK; }
      else                  { fpstat = igFail; }
      break;

    default:
      igAbend(__FILE__, __LINE__, "invalid loop status in igImitate()");
      break;
    }

  return fpstat;
  }


