
/*******************************************************************************

    iggsl.c
    6 Apr 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "iggsl.h"

#include <gsl/gsl_multiroots.h>    /* includes gsl_vector.h */

/*******************************************************************************
This is a wrapper around the user-supplied function to allow it to be used by 
GSL.  To turn a fixed-point problem into a root-finding problem we need to 
subtract X from the result returned by igFunction().  We also need to translate 
the returned status into GSL-speak.  The parameter struct *p is ignored.
*******************************************************************************/

static int gslFunc(const gsl_vector *x, void *p, gsl_vector *f)
  {
  int i;

  if (igFunction(x->data, f->data)) 
    {
    for (i = 0; i < igDim; i++) { f->data[i] -= x->data[i]; }
    return GSL_SUCCESS;
    }

  return GSL_EBADFUNC;
  }

/*******************************************************************************
Given a non-empty igList, this iterates the GSL multiroot solver, starting at 
the last X point in the list, to polish the approx fixed-point.  Up to cnt 
iterations are performed, and we stop if get a within-tolerance FP, hit the 
limit, or have a problem.  The final X/Y/norm triple generated (if any) is 
appended to the list.  If no valid X/Y/norm tuple is appended to the list, the 
call fails.  TBA: Inside the loop, we could finesse the repeated copying of the
X values, since we only need to do this if we've a fixed-point.
*******************************************************************************/

/* A block of code to free some allocated stuff.  Note that freeing xfp and yfp
are guarded (ie, they can be 'turned off' by setting them to NULL). */

#define FREEIT if (xfp != NULL) { free(xfp); } \
               if (yfp != NULL) { free(yfp); } \
               gsl_multiroot_fsolver_free(s);  \
               gsl_vector_free(x);

igFPSTAT igPolish(igList *igl, int cnt)
  {
  const gsl_multiroot_fsolver_type *T;
  gsl_multiroot_fsolver *s;
  gsl_multiroot_function f = {&gslFunc, igDim, NULL};
  gsl_vector *x;

  int iter, i;
  double *xfp = NULL, *yfp = NULL, n;

  if (igl->size < 1 || cnt < 1) { return igFail; }

  T = gsl_multiroot_fsolver_hybrids;
  s = gsl_multiroot_fsolver_alloc(T, igDim);

  x = gsl_vector_alloc(igDim);
  for (i = 0; i < igDim; i++) { x->data[i] = igl->dx[igl->size-1][i]; }

  gsl_multiroot_fsolver_set(s, &f, x);

  for (iter = 0; iter < cnt; iter++)
    {
    if (gsl_multiroot_fsolver_iterate(s)) { break; }

    if (xfp == NULL)               { xfp = igGetVect(igDim); }
    for (i = 0; i < igDim; i++)    { xfp[i] = s->x->data[i]; }
    if (!igVectFinite(igDim, xfp)) { FREEIT  return igFail; }

    if (yfp == NULL)              { yfp = igGetVect(igDim); }
    if (!(*igFunction)(xfp, yfp)) { FREEIT  return igFail; }

    n = (*igNorm)(igDim, xfp, yfp);
    if (!isfinite(n)) { FREEIT  return igFail; }

    if (n < igTol)
      {
      igAppend(igl, xfp, yfp, n);
      xfp = yfp = NULL;
      FREEIT
      return igFP;
      }
    }

  /* We fall through to here if there's a 'problem' or we've hit the iteration
  limit.  Add the latest X/Y/norm triple to the list (if possible), tidy-up and
  return the appropriate status. */

  if (xfp == NULL)               { xfp = igGetVect(igDim); }
  for (i = 0; i < igDim; i++)    { xfp[i] = s->x->data[i]; }
  if (!igVectFinite(igDim, xfp)) { FREEIT  return igFail; }

  if (yfp == NULL)              { yfp = igGetVect(igDim); }
  if (!(*igFunction)(xfp, yfp)) { FREEIT  return igFail; }

  n = (*igNorm)(igDim, xfp, yfp);
  if (!isfinite(n)) { FREEIT  return igFail; }

  igAppend(igl, xfp, yfp, n);
  xfp = yfp = NULL;
  FREEIT
 
  if (n < igTol) { return igFP; }
  return igOK;
  }


