
/*******************************************************************************

    igig.c
    3 Jun 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "igig.h"

#include <stdio.h>
#include <stdlib.h>

int igigAttempts, igigGuesses;
double igigDmin, igigDmax;
int igigLimit1, igigLimit2, igigLimit3;

/* Initialise the system, and set 'reasonable' defaults.  NB: Some of the ig---
system variables (eg, igDim, igFunction, igDmin, igDmax) need to be set 
explicitly for this code to run correctly. */

void igigSetup(void)
  {
  /* From ig.h/c */

  igSeedRand(0U);          /* generate (& save) non-zero random seed */

  /* From igfp.h/c */

  igTol = 1.0E-6;          /* epsilon for FP tolerance */
  igZero = 4.0E-15;        /* floating-point zero tolerance */
  igPLimit = 64;           /* max pivots per Lemke-Howson call */
  igLHPivot = igLHlast;    /* use last X point as initial pivot */
  igNorm = *igLinfNorm;    /* use L-infinity norm for FP tolerance */

  /* From igig.h/c */

  igigAttempts = 10;       /* try to find an FP 10 times */
  igigGuesses = 1000;      /* pick initial X's from 1000 guesses */
  igigLimit1 = 32;         /* IG sizes up to 32x32 allowed */
  igigLimit2 = 64;         /* up to 64 GSL polishing iterations */
  igigLimit3 = 16;         /* up to 16 IG/polish loops */
  }

/*******************************************************************************
This wrapper function tries to find an approx FP of the function.  If we 
succeed, we return igFP and append the X/Y/norm triple to the list.  If there 
are any problems, we return igFail and the list is unchanged.  Otherwise, we 
return igOK and append the best X/Y/norm triple we found to the list.  The input
list must be initialised, but may or may not be empty.  Any existing data on the
input list is unaltered.  If the debug arg is true (ie, any bits are set), then 
various logging info is dumped to the o/p.
*******************************************************************************/

/* A block of code to update the copy of the best approx'n so far, from the last
enry in the working list.  TBA: Make more efficient. */

#define  UPDATE  if (x != NULL) { free(x); } \
                 if (y != NULL) { free(y); } \
                 x = igCopyVect(igDim, wList->dx[wList->size-1]); \
                 y = igCopyVect(igDim, wList->dy[wList->size-1]); \
                 n = wList->dn[wList->size-1];

/* A block of code to manipulate the (non-empty) working list so that it's sole
entry is the best entry. */

#define  MUNGE  nn = wList->dn[0]; \
      jj = 0; \
      for (ii = 1; ii < wList->size; ii++) \
        { if (wList->dn[ii] < nn) { nn = wList->dn[ii];  jj = ii; } } \
      if (jj != 0) \
        { \
        pp = wList->dx[0];  wList->dx[0] = wList->dx[jj];  wList->dx[jj] = pp; \
        pp = wList->dy[0];  wList->dy[0] = wList->dy[jj];  wList->dy[jj] = pp; \
        nn = wList->dn[0];  wList->dn[0] = wList->dn[jj];  wList->dn[jj] = nn; \
        } \
      igShrinkList(wList, 1);

igFPSTAT igigFindFP(igList *igl, unsigned int debug)
  {
  double *x = NULL, *y = NULL, n = -2.0;     /* best approx FP so far */
  int a, l;                                  /* attempt & loop counts */
  igList *wList;                             /* working list */
  igFPSTAT igfps;                            /* status return */
  double ninit;                              /* norm at inner loop entry */

  int ii, jj;                                /* work variables */
  double nn, *pp;                            /* ditto */

  if (igl == NULL) { return igFail; }
  if (debug) { printf("\n"); }

  wList = igNewList(igDim);

  for (a = 0; a < igigAttempts; a++)
    {
    if (debug) { printf("attempt = %d, best norm = %0.15g ...\n", a, n); }

    /* Each attempt starts with the generation of an initial guess.  We need to
    'delete' any existing data on the working list, and we may guess an FP or
    fail. */

    igShrinkList(wList, 0);

    igfps = igGuess(wList, igigGuesses, igigDmin, igigDmax);
 
    if (igfps == igFail) 
      { 
      if (debug) { printf("  fail (guess)\n"); }
      continue;
      }

    if (igfps == igFP)
      {
      UPDATE
      igFreeList(wList);
      igAppend(igl, x, y, n);
      return igFP;
      }

    /* We fall though to here if we have a non-FP guess.  Record the guess if
    it's the best so far, and then loop over the IG/polish methods. */

    if (n < -1.0 || wList->dn[wList->size-1] < n) { UPDATE }

    /* At this point it is guaranteed that the best approx'n so far has been set
    and that the working list is non-empty (with one entry). */

    for (l = 0; l < igigLimit3; l++)
      {
      ninit = wList->dn[wList->size-1];
      if (debug) 
        { printf("  loop = %d, current norm = %0.15g ...\n", 
                                                 l, wList->dn[wList->size-1]); }

      /* Run imitation game phase */

      igfps = igImitate(wList, igigLimit1, 0U); 

      if (igfps == igFail) 
        { 
        if (debug) { printf("    fail (imitate)\n"); }
        break;
        }

      if (igfps == igFP)
        {
        UPDATE
        igFreeList(wList);
        igAppend(igl, x, y, n);
        return igFP;
        }

      MUNGE
      if (wList->dn[wList->size-1] < n) { UPDATE }
      if (debug) 
        { printf("    imitate norm = %0.15g\n", wList->dn[wList->size-1]); }

      /* Run the polish phase */

      igfps = igPolish(wList, igigLimit2);

      if (igfps == igFail) 
        {
        if (debug) { printf("    fail (polish)\n"); }
        break;
        }

      if (igfps == igFP)
        {
        UPDATE
        igFreeList(wList);
        igAppend(igl, x, y, n);
        return igFP;
        }

      MUNGE
      if (wList->dn[wList->size-1] < n) { UPDATE }
      if (debug) 
        { printf("    polish norm = %0.15g\n", wList->dn[wList->size-1]); }

      /* At this point we've done some imitation & some polishing of the approx
      at entry, without failing or finding a FP.  If there's been no progress,
      break out of the loop. */

      if (ninit <= wList->dn[wList->size-1])
        {
        if (debug) { printf("    fail (improvement)\n"); }
        break;
        }
      }    /* end of for (l ...) loop */
    }    /* end of for (a ...) loop */

  /* If we fall through to here, we did not find a within-tolerance FP.  We 
  append the best approx'n (if any) to the list. We need to free any working
  space, but not any appended values. */

  if (wList != NULL) { igFreeList(wList); }

  if (x != NULL) { igAppend(igl, x, y, n);  return igOK; }
  return igFail;
  }


