/*******************************************************************************

    igtableau.c
    22 Apr 11
    Colin Ramsay
    Copyright (C) 2011, The Gambit Project

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "igtableau.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/*******************************************************************************
Allocate, but do not initialise, the memory for the tableau.  A tableau is a 3D 
matrix, representing Dim1 x (2+Dim1+Dim2) and Dim2 x (2+Dim2+Dim1) matrices, 
containing the strategies in the basis (1st col), their weights (2nd col), and 
then the coefficients of the slacks and the variables.  NB: The array indices 
start at 0, but the numbering for slacks/variables starts at 1.
*******************************************************************************/

static double*** allocTableau(int d1, int d2)
  {
  int i;
  double ***t;

  if ((t = (double***)malloc(2 * sizeof(double**))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc t in allocTableau()"); }

  if ((t[0] = (double**)malloc(d1 * sizeof(double*))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc t[0] in allocTableau()"); }

  for (i = 0; i < d1; i++) 
    { 
    if ((t[0][i] = (double*)malloc((2 + d1 + d2) * sizeof(double))) == NULL)
      { igAbend(__FILE__, __LINE__, "can't malloc t[0][i] in allocTableau()"); }
    }

  if ((t[1] = (double**)malloc(d2 * sizeof(double*))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc t[1] in allocTableau()"); }

  for (i = 0; i < d2; i++) 
    { 
    if ((t[1][i] = (double*)malloc((2 + d1 + d2) * sizeof(double))) == NULL)
      { igAbend(__FILE__, __LINE__, "can't malloc t[1][i] in allocTableau()"); }
    }

  return t;
  }

/*******************************************************************************
Builds the two sub-tableaux for the imitation game needed by the algorithm, 
starting from the -||.||^2 values (for player 1) and the identity matrix (for 
player 2).  The format used is that in gambit-lcp, which matches the survey 
paper by McLennan et al.  NB: The initial state of this tableau is specific to 
the imitation game fixed-point algorithm.
*******************************************************************************/

static void 
buildTableau(igTableau *igt, double **game, int dim1, int dim2, double min)
  {
  int i, j;
  double off = 1.0 - min;

  /* The first column represents the variables in the basis (for that row), with
  the convention that a negative number represents the slack variable associated
  with the corresponding positive variable.  The second column is the actual 
  first column of the tableau, and represents, during the execution of the 
  algorithm, the (relative) value of the variable in the basis for that row.
  Initially, the basis is (-1,..,-(d1+d2)), and all the values are -1. */

  for (i = 0; i < dim1; i++)
    {
    igt->tab[0][i][0] = -(double)(i + 1);
    igt->tab[0][i][1] = -1.0;
    }

  for (i = 0; i < dim2; i++) 
    {
    igt->tab[1][i][0] = -(double)(i + dim1 + 1);
    igt->tab[1][i][1] = -1.0;
    }

  /* Slack variables are initial basis, so their columns are the identity. */

  for (i = 0; i < dim1; i++) 
    {
    for (j = 2; j < 2+dim1; j++) 
      { igt->tab[0][i][j] = 0.0; }
    igt->tab[0][i][i+2] = 1.0;
    }

  for (i = 0; i < dim2; i++) 
    {
    for (j =  2; j < 2+dim2; j++) 
      { igt->tab[1][i][j] = 0.0; }
    igt->tab[1][i][i+2] = 1.0;
    }

  /* The initial sub-tableaux are A* & Idim1, with a*_ij = a_ij + 1 - min. */

  for (i = 0; i < dim1; i++) 
    {
    for (j = (2 + dim1); j < (dim1+dim2+2); j++) 
      { igt->tab[0][i][j] = game[i][j - 2 - dim1] + off; }
    }

  for (i = 0; i < dim2; i++) 
    {
    for (j =  (2 + dim2); j < (dim1+dim2+2); j++) 
      { igt->tab[1][i][j] = 0.0; }
    igt->tab[1][i][i+2+dim2] = 1.0;
    }
  }

/* NB: The value of "zero" (ie, "tol") must *not* actually be 0. */

igTableau* igNewTableau(igMatrix *igm, double tol)
  {
  igTableau *igt;

  if ((igt = (igTableau*)malloc(sizeof(igTableau))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igNewTableau()"); }

  /* NB: dim1 = dim2 (= size), since an igMatrix is always square. */

  igt->tab = allocTableau(igm->size, igm->size);
  buildTableau(igt, igm->dpp, igm->size, igm->size, igm->min);

  igt->dim1 = igt->dim2 = igm->size;

  if ((igt->ind = (int*)malloc(igt->dim1 * sizeof(int))) == NULL)
    { igAbend(__FILE__, __LINE__, "can't malloc in igNewTableau()"); }

  igt->numP = -1;            /* -1 flags the "LH never run" state */
  igt->zero = fabs(tol);

  return igt;
  }

/* NB: This frees the struct itself, as well as all its allocated space. */

void igFreeTableau(igTableau *igt)
  {
  int i;

  if (igt == NULL) { return; }

  for (i = 0; i < igt->dim1; i++) { free(igt->tab[0][i]); }
  free(igt->tab[0]);

  for (i = 0; i < igt->dim2; i++) { free(igt->tab[1][i]); }
  free(igt->tab[1]);

  free(igt->tab);
  free(igt->ind);

  free(igt);
  }

/*******************************************************************************
Tells if strategy "strategy" (which may be a real strategy, +ve, or a slack 
variable, -ve) is in basis of the current tableau.  If not, it returns the same 
strategy, since we can pivot on it.  If it's already in the base, we can't pivot
on it but we can on its 'complement'.  Note that the basis is always 
complementary (at a NE) or almost-complementary (during execution of 
Lemke-Howson), and we (should) only call this when the tableau represents a NE 
(ie, at entry to igLemHow()).  NB: The strategy value should be in the range 
(+/-1 ... +/-dim1+dim2). 
*******************************************************************************/

static int getPivot(double*** t, int dim1, int dim2, int strategy) 
  {
  int i;

  for (i = 0; i < dim1; i++)
   { if (t[0][i][0] == strategy) { return -strategy; } }

  for (i = 0; i < dim2; i++) 
    { if (t[1][i][0] == strategy) { return -strategy; } }

  return strategy;
  }

/*******************************************************************************
Determines which of the two sub-tableaux is to be used for pivoting; this is the
one which contains the COLUMN for the strategy (real or slack).  We assume that 
the strategy value is in the correct range, and don't check for errors.
*******************************************************************************/

static int getTableau(int dim1, int dim2, int strategy)
  {
  if (strategy > dim1 || (strategy < 0 && strategy >= -dim1))
    { return 0; }
  return 1;
  }

/*******************************************************************************
Returns the column index in the tableau corresponding to pivoting on the given 
strategy (cf, getTableau()).  We assume that the strategy value is always in the
correct range, and don't check for errors. 
*******************************************************************************/

static int getColumn(int dim1, int dim2, int strategy) 
  {
  if (strategy > 0)
    {
    if (strategy > dim1) { return (1 + strategy); }
    return (1 + dim2 + strategy);
    }

  if (strategy >= -dim1) { return (1 - strategy); }
  return (1 - strategy - dim1);
  }

/*******************************************************************************
Runs the Lemke-Howson algorithm (aka, LHA) on the tableau, 'moving' it from the 
starting Nash equilibrium (aka, NE) to another NE.  The starting NE is usually 
the null (ie, artificial) one, but it can be (the tableau for) any one.  "initP"
is the initial pivot to use, and "limP" is the limit on the number of pivots to
apply.  The returned value gives the result of the LHA; a NE or some sort of 
problem.  The strategy vectors for the NE (if any) can be extracted by the 
igGet*Prob() routines, and the number of pivots by igGetPivots().

Although the LHA is general purpose, the igTableau structure is designed for the
imitation game fixed-point algorithm, where the LHA is run once on each game.
We also can't (re)run the LHA if a previous run did not generate a valid NE.  To
prevent multiple runs, the "numP" flag value of -1 is used.  In principle, if a
particular starting pivot did not generate a valid NE, another run of the LHA 
with a different starting pivot might, but there is no support for this (other
than building a new igTableau).
 
There are potential numerical problems here due to the continual pivoting of the
tableau.  (There are always at least 2 pivot steps, and perhaps an exponential 
number.)  We try to mitigate these by using the zero value as a tolerance when 
doing comparisons (with 0), and by forcing 'approx' value to exactly 0 or 1.  If
we detect a problem, we return an error status.

Since (normalised) floats have ~15 decimal digits of precision, and the affine
transformation applied to the game sets the minimum tableau value to 1, zero 
values of ~10^-14 seem appropriate.  The most common 'problem' seems to be 
values close to 0 (due to, eg, imperfect cancellation when subtracting?); these 
can be forced to 0 by a suitable choice of zero.

All of the above means that: finding a NE is not guaranteed; even if one is 
found it may have an all-0 probability vector or be only 'approximate'.
*******************************************************************************/

igLHSTAT igLemHow(igTableau *igt, int initP, int limP)
  {
  int newpivot;
  double agg, coeff;
  int i, j;
  int pivot;
  int ntab, nlines, column;
  double **tab;    

  /* max is the current maximum ratio seen so far, while val is the current
  value.  igt->ind keeps track of the row indices where the max occurs.  rcnt is
  the number of times the max was seen; ditto rnct2.  index is the row index 
  (ie, ind[0]) to be used, while row is the current row (aka ind[i]). */

  double max, val, vm;
  int rcnt, index, rcnt2, row;

  /* Check the arguments and the status, and record function entry. */

  if (igt->numP != -1) { return igNotInit; }
  if (initP < 1 || initP > igt->dim1 + igt->dim2) { return igBadArg; }
  if (limP < 2) { return igBadArg; }

  igt->numP = 0;

  /* initP is the "label" of the variable we want to pivot on.  getPivot()
  determines, looking at the tableau, if we want the real strategy (+ve value)
  to enter the basis, or the corresponding complementary variable (-ve value).  
  This happens, for example, when we don't start pivoting from the artificial 
  equilibrium, but from an actual one, having in its basis some real strategies.
  If we pivot on them, we want the complementary variable to enter the basis. */
  
  pivot = getPivot(igt->tab, igt->dim1, igt->dim2, initP);

  while (igTRUE)
    {
    (igt->numP)++;
    if (igt->numP > limP) { return igPivLim; }

    /* ntab is the tableau we are in (0 or 1).  nlines is the number of rows of 
    the tableau we are in (dim1 or dim2).  column is the tableau column which 
    corresponds to the variable we are pivoting on.  tab is shorthand for 
    igt->tab[ntab]. */

    ntab = getTableau(igt->dim1, igt->dim2, pivot);
    nlines = (ntab == 0 ? igt->dim1 : igt->dim2);
    column = getColumn(igt->dim1, igt->dim2, pivot);
    tab = igt->tab[ntab];

    /* We choose among the rows of the current tableau for which the coefficient
    of the variable entering the basis is > 0 (if it's <= 0 we cannot choose 
    this row).  We use the maximum ratio test to select the row.  We choose the
    index so that we max the ratio between the weight of the variable in basis 
    (ie, tab[ntab][i][1]) and the coefficient of the variable entering (ie,
    tab[ntab][i][column]; NB: this is -ve).  If this does not yield a unique row
    (ie, the system is degenerate) we extend the ratio test to the other columns
    and use the lexico-maximal row. */

    /* Set things up, and do the first pass.  Note that the value of max here
    is actually irrelevant, since it'll be set the first time through by the
    the rcnt == 0 test. */

    rcnt = 0;
    max = 0.0;

    for (i = 0; i < nlines; i++) 
      {
      if (tab[i][column] < igt->zero)      /* Not positive enough */
	{ continue; }
      
      val = tab[i][1] / tab[i][column];    /* Ratio */

      vm = val - max;
      if (rcnt == 0 || vm > igt->zero) 
        { 
        /* If this is the first time through, or if we have a larger ratio, we 
        set the row count and the maximum value and store the row index i. */ 

	max = val;
        rcnt = 1;
	igt->ind[0] = i;
        } 
      else if (vm >= -(igt->zero))
        {
        /* If we repeat a maximum, we need to update the data structure. */

        igt->ind[rcnt] = i;
        rcnt++;
        }
      }

    /* If we didn't update the rcnt value, this means there isn't a row for 
    which the coefficient of the variable entering the basis is greater than 
    zero.  If we are in this condition, something's gone awry. */

    if (rcnt < 1) { return igNoRow; }

    /* There are one or more maximum ratio rows.  If there's just one, we're
    done.  If not, the system is degenerate.  One way to cope with degenerate 
    systems (ie, to avoid cycling) is to iterate over the columns maintaining a
    set of the lexico-maximal ones and stop when this is a singleton. */

    if (rcnt > 1)
      {
      /* Repeat the code above for the slack columns (ex the pivot column), 
      keeping track of the number of max ratios until this drops to 1. */

      while (rcnt > 1)
        {
        for (j = 2; j <= nlines+1; j++)    /* for all slacks columns ... */
          {
          if (j == column)               /* NB: It's not unusual to hit this */ 
            { continue; }

          /* Same as the above for the pivot column.  We work through the old
          rnct rows in ind[], accumulating the new rows (a subset of the old) in
          positions 0..rcnt2-1. */

          rcnt2 = 0;
          max = 0.0;

          for (i = 0; i < rcnt; i++) 
            {
            row = igt->ind[i];
            val = tab[row][j] / tab[row][column];

            vm = val - max;
            if (rcnt2 == 0 || vm > igt->zero) 
              { 
	      max = val;
              rcnt2 = 1;
              igt->ind[0] = row;
              } 
            else if (vm >= -(igt->zero))
              {
              igt->ind[rcnt2] = row;
              rcnt2++;
              }
            }

          rcnt = rcnt2;
          if (rcnt == 1) { break; }
          }

        if (j > nlines+1) { return igLexMax; }
        }
      }

    /* Finally we choose what variable will go out of the basis */
 
    index = igt->ind[0];
    newpivot = (int)tab[index][0];
    
    /* Now we know what variable will leave the basis, we only need to do two
    things: solve the equation we chose with the maximum ratio test, updating 
    the variable in the basis; and solve all other equations of the tableau, 
    updating all the coefficients. */

    /*  In the first step we: update the row chosen with the maximum ratio test;
    update tab[ntab][index][0], which tells what variable is in basis; calculate
    the coefficient we will divide all other coefficients with; update the whole
    row; and set the coefficient of the variable entering the basis to zero. */
    
    tab[index][0] = pivot;    /* NB: Puts pivot in & newpivot out of basis */

    coeff = tab[index][column];
    for (i = 1; i < (igt->dim1 + igt->dim2 + 2); i++)
      { tab[index][i] /= coeff; }

    tab[index][column] = 1.0;    /* Ensure this is *exactly* one. */

    /* In the second step, we solve all the other equations in the tableau by
    subtracting suitable multiples of the pivot row.  We check if the 
    coefficient of the variable entering in basis in this row is nonzero.  If 
    so, we update the coefficients, and set to zero the coefficient of the 
    variable entering the basis.  NB: the check for "incomplete" cancellation to
    zero in all columns (incl [1]). */
    
    for (i = 0; i < nlines; i++) 
      {
      if (i == index) { continue; }

      agg = tab[i][column];

      if (agg < -(igt->zero) || agg > igt->zero) 
        {
	for (j = 1; j < (igt->dim1 + igt->dim2 + 2); j++) 
          {
	  tab[i][j] -= agg * tab[index][j];

          if (-(igt->zero) < tab[i][j] && tab[i][j] < igt->zero 
                                                            && tab[i][j] != 0.0)
            { tab[i][j] = 0.0; }
	  }
        }
      else     /* ignore row */
        { ; }

      tab[i][column] = 0.0;    /* Ensure this is *exactly* zero. */
      }

    /* In principle, column[1] (ie, the basis weights) should be <= 0.  Due to
    numerical problems, it may be > 0.  This can cause problems.  So we check 
    for this, and force the value to 0.0. */

    for (i = 0; i < nlines; i++) 
      {
      val = tab[i][1];
      if (val > -(igt->zero) && val != 0.0)
        { tab[i][1] = 0.0; }
      }

    /* Following the complementary pivoting rule, the new variable to pivot on 
    is the complement of the old variable. */

    pivot = -newpivot;
    
    /* We stop the execution of the algorithm when either we pivot on the first 
    variable we pivoted on (that strategy is leaving the basis), or on the 
    complement of that variable (that complement is leaving the basis).  In both
    cases, we are not in a k-almost complete equilibrium, but in an actual Nash 
    equilibrium. */
    
    if (newpivot == initP || newpivot == -initP) 
      { break; }
    }

  return igNE;
  }

/*******************************************************************************
To extract the X or Y (player 1 or 2) probability vectors, we need to normalize 
the vector of the strategy (remembering that the weights are -ve), to obtain a 
sum of 1 for the probabilities.  Thus, we need to accumulate column[1]'s which 
have +ve column[0]'s and so represent variables (not slacks).  The column[1] 
values should be "< -zero" or "= 0.0" and the returned probabilities should be 
"> +zero" or "= 0.0", but there may be numerical problems and all this has to be
checked and 'adjusted'.  If there are any problems, a NULL is returned.

The 1st sub-tableau (ie, tab[0]) has as its basis the slacks for player 1 and 
the strategies for player 2.  Thus the probabilities for the Y (ie, 2nd) player 
are in the tab[0][i][1] column, and vice versa for player X.  The variables are 
*not* in order in the 0th column.

Notes: The X & Y probability vectors do *not* have the same length as the 
dimension of the fixed-point function.  It's the user's responsibility to free 
the returned vector (if any).  The normalising value "tot" is the reciprocal of 
the expected payoffs (for the tableau, *not* necessarily the original game). 
For imitation games, X's probability vector has equal values for all the non-
zero entries.
*******************************************************************************/

double* igGetXProb(igTableau *igt)
  {
  int i, v;
  double tot = 0.0, w, *xp;

  xp = igGetCVect(igt->dim1, 0.0);

  for (i = 0; i < igt->dim2; i++) 
    {
    v = igt->tab[1][i][0];     
    if (v > 0)                 /* strategy labels are 1 .. dim1 */
      {
      w = -(igt->tab[1][i][1]);
      if (w > 0.0) 
        { xp[v - 1] = w;  tot += w; }
      }
    }

  if (tot < igt->zero)
    { free(xp);  return NULL; }

  for (i = 0; i < igt->dim1; i++) 
    { if (xp[i] > 0.0) { xp[i] /= tot; } }

  return xp;
  }

double* igGetYProb(igTableau *igt)
  {
  int i, v;
  double tot = 0.0, w, *yp;

  yp = igGetCVect(igt->dim2, 0.0);

  for (i = 0; i < igt->dim1; i++) 
    {
    v = igt->tab[0][i][0];     
    if (v > 0)                 /* strategy labels are dim1+1 .. dim1+dim2 */
      {
      w = -(igt->tab[0][i][1]);
      if (w > 0.0) 
        { yp[v - 1 - igt->dim1] = w;  tot += w; }
      }
    }

  if (tot < igt->zero)
    { free(yp);  return NULL; }

  for (i = 0; i < igt->dim2; i++) 
    { if (yp[i] > 0.0) { yp[i] /= tot; } }

  return yp;
  }

int igGetPivots(igTableau *igt)
  { return igt->numP; }

/* Does a basic printout of the two sub-tableaux. */

void igDumpTab(igTableau *igt)
  {
  int i, j;

  for (i = 0; i < igt->dim1; i++) 
    {
    printf("  %2d", (int)(igt->tab[0][i][0]));
    printf("  %0.6f ", igt->tab[0][i][1]);
    for (j = 2; j < (2 + igt->dim1 + igt->dim2); j++) 
      { printf(" %0.6f", igt->tab[0][i][j]); }
    printf("\n");
    }

  for (i = 0; i < igt->dim2; i++) 
    {
    printf("  %2d", (int)(igt->tab[1][i][0]));
    printf("  %0.6f ", igt->tab[1][i][1]);
    for (j = 2; j < (2 + igt->dim1 + igt->dim2); j++) 
      { printf(" %0.6f", igt->tab[1][i][j]); }
    printf("\n");
    }
  }


